"""
File Purpose: single-fluid mhd bases loader
"""
import numpy as np

from ..defaults import DEFAULTS
from ..dimensions import SINGLE_FLUID
from ..errors import (
    FluidValueError, LoadingNotImplementedError, SetvarNotImplementedError,
    InputError, InputMissingError,
)
from ..quantities import AllBasesLoader, SimpleDerivedLoader
from ..tools import simple_property


''' --------------------- MhdBasesLoader --------------------- '''

class MhdBasesLoader(AllBasesLoader, SimpleDerivedLoader):
    '''base quantities based on single-fluid mhd variables: r, e, u, and B;
        r = mass density.
            Subclass might know n (number density) and m (mass) instead,
                in which case it should override get_r, get_n and get_m.
        e = internal energy density.
            Subclass might know P (pressure) or T (temperature) instead,
                in which case it should override get_e, get_P, and get_T.
        u = velocity (vector).
            Subclass might know p (momentum density) instead,
                in which case it should override get_u and get_p.
            (Subclasses knowing u already get SimpleDerivedLoader.get_p() == u * r.)
        B = magnetic field (vector).
    '''

    # # # SINGLE FLUID MODE CHECK # # #
    @property
    def in_single_fluid_mode(self):
        '''whether self is in "single fluid mode".
        MhdCalculator is always in single-fluid mode... but subclass might not be.
        Some vars (bases especially) assume single-fluid mode, so it is good to check.
            E.g. this ensures multifluid subclasses don't do weird things by accident.
        '''
        try:
            return self.fluid is SINGLE_FLUID
        except AttributeError:  # self doesn't have 'fluid' --> definitely single fluid mode.
            return True

    def assert_single_fluid_mode(self, varname='var', *, mode='getting'):
        '''asserts that self is in single fluid mode; else crash.
        varname: str
            name of var to include in error message if crashing.
        mode: str, 'getting' or 'setting'
            determines error type & error message if crashing.
            'getting' --> FluidValueError, with error message starting like:
                "{type(self)} getting {varname} requires self.in_single_fluid_mode..."
            'setting' --> SetvarNotImplementedError, with error message starting like:
                "var={varname}, when self not in_single_fluid_mode..."

        Use for operations which directly assume single fluid mode (e.g. get_r implementation here).
        Not necessary for operations which apply regardless of number of fluids,
            E.g. n = r / m for any number of fluids, and B is always independent of fluid.
        If unsure, err on the side of caution and use this function,
            to require multifluid subclass to explicitly handle the situation else crash.
        '''
        if not self.in_single_fluid_mode:
            if mode == 'getting':
                errmsg = (f'{type(self).__name__} getting {varname!r} requires self.in_single_fluid_mode,\n'
                          f'but got self.fluid={self.fluid!r} instead of SINGLE_FLUID.')
                raise FluidValueError(errmsg)
            elif mode == 'setting':
                errmsg = (f'var={varname!r}, when self not in_single_fluid_mode;\n'
                          f'got self.fluid={self.fluid!r} instead of SINGLE_FLUID.')
                raise SetvarNotImplementedError(errmsg)
            else:
                raise InputError(f"invalid mode={mode!r}; expected 'getting' or 'setting'.")

    @known_pattern(r'(SF|SINGLE_FLUID)_(.+)', deps=[1])  # SF_{var} or SINGLE_FLUID_{var}
    def get_single_fluid_var(self, var, *, _match=None):
        '''SF_{var} (or SINGLE_FLUID_{var}) --> {var}, definitely in single fluid mode.
        crashes with FluidValueError if not self.in_single_fluid_mode.

        The implementation here just does self.assert_single_fluid_mode(),
            then returns self(var) (var from SF_var or SINGLE_FLUID_var string).
        Subclass might override to also set self.fluid = SINGLE_FLUID.
        '''
        _sf, here, = _match.groups()
        self.assert_single_fluid_mode(here)
        return self(here)

    # # # BASES # # #
    @known_var
    def get_ne(self):
        '''electron number density
        [Not implemented for this class]
        '''
        raise LoadingNotImplementedError('ne')

    # # # SIMPLE DERIVED # # #
    @known_var(deps=['e', 'r'])
    def get_eperm(self):
        '''internal energy (total, not density) per unit mass. eperm = e / r.'''
        return self('e') / self('r')

    @known_var(deps=['curl_B'])
    def get_J(self):
        '''current density (without displacement current). J = curl(B) / mu0.
        Per unit area, e.g. the SI units would be Amperes / meter^2.
        '''
        return self('curl_B') / self.u('mu0')

    # # # DERIVED FROM ASSUMED ABUNDANCES # # #
    cls_behavior_attrs.register('elements')
    elements = simple_property('_elements', setdefaultvia='_get_default_elements',
        doc='''ElementList of all elements in multifluid mixture assumed from single-fluid mhd.
        Used to infer SINGLE_FLUID m.''')

    def _get_default_elements(self):
        '''default value for elements. here, just returns self.tabin.elements,
        but raises a helpful error message if self.tabin not set.
        '''
        try:
            tabin = self.tabin
        except (AttributeError, InputMissingError):
            errmsg = (f"{type(self).__name__}.elements not set, and can't get default when self.tabin missing.\n"
                      "Either set self.elements or self.tabin, to proceed.")
            raise InputMissingError(errmsg)  # not from None; we want to keep tabin error info.
        return tabin.elements

    @known_var
    def get_m(self):
        '''abundance-weighted average fluid particle mass.
        m = self.elements.mtot() * (mass of 1 atomic mass unit).
        The "abundance-weighting" is as follows:
            m = sum_x(mx ax) / sum_x(ax), where ax = nx / nH, and x is any elem from self.elements.
            note: ax is related to abundance Ax via Ax = 12 + log10(ax).
        see help(self.elements.mtot) for more details, including a proof that mtot = rtot / ntot.

        (if self.elements does not exist, this method will crash with LoadingNotImplementedError)
        '''
        self.assert_single_fluid_mode('m')
        try:
            elements = self.elements
        except AttributeError:
            errmsg = f'{type(self).__name__}.get_m() when self.elements does not exist'
            raise LoadingNotImplementedError(errmsg) from None
        return elements.mtot() * self.u('amu')

    @known_var(deps=['r', 'm'])
    def get_n(self):
        '''number density. n = (r / m) = (mass density / mass)
        (Note: for single-fluid, this excludes electron number density.)
        '''
        r = self('r')
        m = self('m')
        r = self._upcast_if_max_n_requires_float64(r)
        return r / m

    def _max_n_requires_float64(self):
        '''returns whether max supported n, in self.units unit system, is too big for float32.
        max supported n, in SI units, is DEFAULTS.MHD_MAX_SAFE_N_SI
        crash with ValueError if max supported n is too big for float64 as well.
        without this check (and conversion to float64 if needed), some n values might become inf.
        '''
        max_n = DEFAULTS.MHD_MAX_SAFE_N_SI * self.u('n', convert_from='si')
        if max_n > np.finfo(np.float32).max:
            if max_n > np.finfo(np.float64).max:
                errmsg = (f'cannot guarantee support for DEFAULTS.MHD_MAX_SAFE_N_SI,'
                          f'{DEFAULTS.MHD_MAX_SAFE_N_SI}, even using float64.')
                raise ValueError(errmsg)
            return True
        return False

    def _upcast_if_max_n_requires_float64(self, array):
        '''if max n requires float64, upcast array to float64,
        unless array's dtype was already float64 or larger (in which case, just return array).
        see self._max_n_requires_float64 for details on when max n requires float64.
        '''
        if self._max_n_requires_float64():
            if np.finfo(array.dtype).max < np.finfo(np.float64).max:
                return array.astype(np.float64)
        else:
            return array

    # # # MISC # # #
    @known_var(deps=['ne', 'SF_n'])
    def get_ionfrac(self):
        '''ionization fraction. ionfrac = ne / n (from single fluid).
        Assumes quasineutrality, and that only once-ionized ions are relevant.
        '''
        self.assert_single_fluid_mode('ionfrac')
        return self('ne') / self('SF_n')

    # # # NEUTRALS # # #
    @known_var(deps=['SF_T'], aliases=['T_n'])
    def get_T_neutral(self):
        '''temperature of neutrals; T_n = T of SINGLE_FLUID.
        (subclass might implement better T, but here assumes T_n equivalent to SF_T.)
        '''
        return self('SF_T')


    # # # --- SETTING VALUES; KNOWN SETTERS --- # # #
    # used when using set_var.

    @known_setter
    def set_r(self, value, **kw):
        '''set r to this value. r = mass density.'''
        return self.set_var_internal('r', value, ['snap', 'fluid'], **kw, ukey='mass_density')

    @known_setter
    def set_n(self, value, **kw):
        '''set n to this value. n = number density == r / m.'''
        self.assert_single_fluid_mode('n', mode='setting')
        # ^^ implementation here doesn't work for mhd multifluid densities,
        #    because the multifluid densities usually infer r from n,
        #    instead of inferring n from r like single-fluid mode does.
        r = value * self('m')
        return self.set('r', r, **kw)

    @known_setter
    def set_e(self, value, **kw):
        '''set e to this value. e = energy density.'''
        self.assert_single_fluid_mode('e', mode='setting')
        return self.set_var_internal('e', value, ['snap', 'fluid'], **kw, ukey='energy_density')

    @known_setter
    def set_eperm(self, value, **kw):
        '''set eperm to this value. eperm = internal energy per unit mass == e / r.
        Depends on the current value of r; if also setting r be sure to set r first.
        '''
        self.assert_single_fluid_mode('eperm', mode='setting')
        e = value * self('r')
        return self.set('e', e, **kw)

    @known_setter
    def set_T_fromtable(self, value, **kw):
        '''set T_fromtable to this value. T_fromtable = single fluid temperature, from er table.
        Depends on the current value of r; if also setting r be sure to set r first.
        (internally, sets eperm such that T is the given value when doing lookups.)
        '''
        with self.using(coords_units=self.coords_units_explicit, units='raw'):
            r = self('SF_r')
        eperm = self.tabin['T'].interp_inverse(value, r=r)
        with self.using(units='raw'):
            return self.set('eperm', eperm, **kw)



"""
File Purpose: tools related to masking xarray objects

for simply filling masked regions with nan, can use xarray_obj.where(...).
But, that can cause a lot of memory full of nan which is not desireable.
Methods in here help with stacking/unstacking mask dimensions,
    so it can be possible to keep only the non-nan points.

Datasets are able to store the full mask info as a data_var,
    so it is possible to ds.pc.mask() or ds.pc.unmask() if '_mask' already in ds.
DataArrays can't store extra dimensions,
    so mask needs to be passed, e.g. arr.pc.mask(mask), arr.pc.unmask(mask)
"""
import numpy as np
import xarray as xr

from .xarray_accessors import pcAccessor
from .xarray_coords import xarray_dims_coords
from .xarray_dimensions import (
    xarray_ensure_dims, _paramdocs_ensure_dims,
    xarray_drop_vars,
)
from .xarray_misc import xarray_as_array, xarray_convert_types
from .xarray_indexing import xarray_where
from ..docs_tools import format_docstring
from ...errors import (
    InputError, InputConflictError, InputMissingError,
    DimensionKeyError,
)


''' --------------------- Masking --------------------- '''

@pcAccessor.register('mask')
@format_docstring(**_paramdocs_ensure_dims)
def xarray_mask(array, mask=None, stackdim='_mask_stack', *, stack=True, create_index=True,
                skip_arrays_without_mask_dims=True, store_mask=None, promote_dims_if_needed=True):
    '''mask an xarray object; result has non-nan values only where mask=False.
    By default, also stack along mask dimensions, and drop points where mask=True.

    array: xarray.DataArray or xarray.Dataset
        object to be masked.
    mask: None or xarray.DataArray
        mask to apply to array.
        None --> use array['_mask'] (and array must be a Dataset)
    stackdim: str, default '_mask_stack'
        result.dims[stackdim] has the mask dimensions stacked.
    stack: bool
        whether to stack result.
        if False, don't stack, just use array.where(~mask). (~ means negation)
    create_index: bool
        whether to create a MultiIndex for stackdim in the result, if stack.
    skip_arrays_without_mask_dims: bool
        whether to apply mask to arrays without mask dims.
        if True, only arrays originally containing at least one mask dim will be masked.
        if False, all results will have mask dims (or stackdim) via array.where(~mask).
    store_mask: None or bool
        whether to store full mask as a data_var in result.
        None --> True if array is a Dataset, else False
        True --> result Dataset will have full mask in data_var '_mask'.
                If stack, mask dims replaced by '_mask_{{d}}' for d in mask.dims,
                    to avoid conflict with original mask dims' coords in result.
                If input DataArray, first convert to Dataset with data_var array.name.
        False --> do not store mask in result.
    promote_dims_if_needed: {promote_dims_if_needed}

    crash with DimensionKeyError if any relevant dim (stackdim, '_mask', '_mask_{{d}}') already exists.
    '''
    if mask is None:
        if xarray_has_mask(array):
            mask, array = xarray_popped_mask(array)
        else:
            raise InputMissingError('mask() requires mask input, or array=<Dataset with "_mask" data_var>.')
    if store_mask is None:
        store_mask = isinstance(array, xr.Dataset)
    if store_mask and isinstance(array, xr.DataArray) and array.name is None:
        raise InputConflictError('store_mask=True when array is a DataArray with name=None.')
    kw_where = dict(skip_if_no_matching_dims=skip_arrays_without_mask_dims)
    if stack:
        array = xarray_ensure_dims(array, mask.dims, promote_dims_if_needed=promote_dims_if_needed)
        stacked = array.stack({stackdim: mask.dims}, create_index=create_index)
        keeping = ~mask.stack({stackdim: mask.dims}, create_index=create_index)  # ~ means negation
        result = xarray_where(stacked, keeping, drop=True, **kw_where)
    else:
        result = xarray_where(array, ~mask, **kw_where)
    if store_mask:
        result = xarray_store_mask(result, mask)
    return result


''' --------------------- Helper methods --------------------- '''

@pcAccessor.register('has_mask')
def xarray_has_mask(array):
    '''returns whether an xarray object has a mask. I.e., has '_mask' data_var.
    (Always False for DataArrays).
    '''
    return isinstance(array, xr.Dataset) and '_mask' in array

@pcAccessor.register('store_mask')
def xarray_store_mask(array, mask):
    '''return Dataset like input but containing data_var '_mask'.
    crash with DimensionKeyError if '_mask' already exists in input.

    if mask.dims not already in array.dims, rename mask dims d to '_mask_{d}'.
        (this will occur when storing mask in a stacked result,
        because the stacked result will have stackdim instead of mask dims.)

    array: xarray.DataArray or xarray.Dataset
        if DataArray, will be converted to Dataset with data_var array.name.
    '''
    if isinstance(array, xr.DataArray):
        if array.name is None:
            raise InputError('cannot store mask in DataArray with name=None.')
        ds = array.to_dataset()
    else:
        ds = array  # input was a Dataset already
    if '_mask' in ds:
        raise DimensionKeyError('dataset already has "_mask"; cannot store mask.')
    if any(d not in array.dims for d in mask.dims):
        mask = mask.rename({d: f'_mask_{d}' for d in mask.dims})
    return ds.assign(_mask=mask)

@pcAccessor.register('stored_mask', totype='dataset')
def xarray_stored_mask(ds):
    '''returns original mask, based on info stored in ds.
    original mask must be stored in data_var '_mask'.
    if ds['_mask'] dims look like '_mask_{d}', rename them to {d}.
    '''
    if '_mask' not in ds:
        if isinstance(ds, xr.DataArray):
            raise InputError("cannot retrieve stored_mask; ds doesn't have '_mask' data_var.")
        else:
            raise InputError(f'cannot retrieve stored_mask from non-Dataset object of type {type(ds)}')
    mask = ds['_mask']
    to_rename = {d: d[len('_mask_'):] for d in mask.dims if d.startswith('_mask_')}
    if to_rename:
        mask = mask.rename(to_rename)
    return mask

@pcAccessor.register('popped_mask', totype='dataset')
def xarray_popped_mask(ds):
    '''returns (stored mask, copy of ds without '_mask')
    See also: xarray_stored_mask, xarray_popped.
    '''
    mask = xarray_stored_mask(ds)
    return mask, xarray_drop_vars(ds, '_mask')


''' --------------------- Unmasking --------------------- '''

_paramdocs_unmask = {
    'stackdim': '''str, default '_mask_stack'
        dimension along which the mask stacking occurred.''',
    'store_mask': '''None or bool
        whether to store full mask as a data_var in result.
        None --> True if result would otherwise be a Dataset, else False
        True --> result will be a Dataset with full mask in data_var '_mask'.
        False --> do not store mask in result.''',
    'as_array': '''None or bool
        whether to ensure result is a DataArray.
        None --> True if result would be a Dataset with a single data_var, else False.
        True --> result will be a DataArray; crash if not possible
                    (e.g. crash if output would have multiple vars, or if store_mask=True).
        False --> result will be a Dataset, unless input was a DataArray and store_mask=False.''',
    'reindex': '''bool
        whether to result.reindex_like(mask). Highly recommended, but not required...''',
    '_upcast_bool': '''bool
        whether to upcast dtype=bool array or data_vars to int8 before unstacking.
        when False, unstacking produces a dtype=object array due to the nans for missing values.
        (when True, unstacking makes dtype=float32, using 0 for False, 1 for True, nan for nan.)''',
}


@pcAccessor.register('unmask')
@format_docstring(**_paramdocs_unmask)
def xarray_unmask(array, mask=None, stackdim='_mask_stack', *, store_mask=False,
                  as_array=None, reindex=True, _upcast_bool=True):
    '''unmask (i.e., unstack) a masked (and stacked) xarray object.
    array.pc.unmask(mask) is equivalent to mask.pc.demask(array).
    See also: xarray_unmask_var, to get a single unmasked var from a Dataset.

    array: xarray.DataArray or xarray.Dataset
        object to unmask.
    mask: None or xarray.DataArray
        the (unstacked) mask. If None, use mask stored in array['_mask'] (and array must be a Dataset).
    stackdim: {stackdim}
    store_mask: {store_mask}
    as_array: {as_array}
    reindex: {reindex}
    _upcast_bool: {_upcast_bool}

    In the simplest case (mask not None; array=non-boolean DataArray with MultiIndex in stackdim),
        this method behaves just like array.unstack(stackdim).reindex_like(mask).
    All the other stuff here helps to handle more complicated cases,
        e.g. Dataset containing mask, possibly without MultiIndex along stackdim.
    '''
    if isinstance(mask, str):
        raise InputError('did not expect mask to be str... maybe you input args in the wrong order?')
    if not isinstance(stackdim, str):
        raise InputError(f'expected str stackdim; got object of type {type(stackdim)}.')
    if stackdim not in array.dims:
        errmsg = f'stackdim={stackdim!r} not found in array.dims={array.dims}.'
        if stackdim in mask.dims:
            errmsg = errmsg + ('\nIf you wanted syntax like mask.pc.unmask(masked_array), '
                               'consider mask.pc.demask(masked_array) instead.')
        raise DimensionKeyError(errmsg)
    # bookkeeping for mask input
    array_has_mask = xarray_has_mask(array)
    if (not array_has_mask) and (mask is None):
        raise InputMissingError("unmask() expects mask kwarg or input array=<Dataset with '_mask' data_var>.")
    if array_has_mask:
        arrmask, array = xarray_popped_mask(array)
        if mask is None:
            mask = arrmask
        elif not np.array_equal(arrmask, mask):
            raise InputConflictError("unmask(array, mask); array['_mask'] conflicts with input mask.")
    # infer stackdim coords from mask if necessary (i.e. if stackdim doesn't have coord already)
    if stackdim not in array.coords:  # need to define MultiIndex coord for stackdim before we can unstack.
        stackdim_coords = xarray_dims_coords(array, include_dims_as_coords=True)[stackdim]
        for mask_d in mask.dims:
            if mask_d not in stackdim_coords:
                errmsg = (f'masked array has no {mask_d!r} coord along stackdim(={stackdim!r})).\n'
                          f'The {mask_d!r} coord is required because it appears in mask.dims={mask.dims}.')
                if array_has_mask: raise DimensionKeyError(errmsg)
                else: raise InputConflictError(errmsg)
        array = array.set_xindex(mask.dims)  # sets stackdim coord
        # (knows about stackdim because mask.dims coords in array are coords along stackdim.)
    if _upcast_bool:
        array = xarray_convert_types(array, {bool: np.int8})
    # >> actually unstack & reindex <<
    result = array.unstack(stackdim)
    if reindex:
        result = result.reindex_like(mask)
    # postprocessing
    if as_array is None and store_mask is None:
        as_array = isinstance(result, xr.DataArray) or len(result.data_vars)==1
        store_mask = (not as_array) and isinstance(result, xr.Dataset)
    if store_mask is None:
        store_mask = isinstance(array, xr.Dataset)
    if store_mask:
        result = xarray_store_mask(result, mask)
    # convert to DataArray if desired & possible
    if as_array is None:
        as_array = isinstance(result, xr.DataArray) or len(result.data_vars)==1
    if as_array:
        result = xarray_as_array(result)
    return result

@pcAccessor.register('demask', totype='array')
@format_docstring(**_paramdocs_unmask)
def xarray_demask_from_mask(mask, array, stackdim='_mask_stack', *, store_mask=False,
                            as_array=None, reindex=True, _upcast_bool=True):
    '''unmask (i.e., unstack) a masked (and stacked) xarray object, using this mask.
    Equivalent to xarray_unmask(array, mask=mask, ...)

    mask: xarray.DataArray
        the (unstacked) mask.
    array: xarray.DataArray or xarray.Dataset
        object to unmask.
    stackdim: {stackdim}
    store_mask: {store_mask}
    as_array: {as_array}
    reindex: {reindex}
    _upcast_bool: {_upcast_bool}
    '''
    return xarray_unmask(array, mask=mask, stackdim=stackdim, store_mask=store_mask,
                         as_array=as_array, reindex=reindex, _upcast_bool=_upcast_bool)

@pcAccessor.register('demask', totype='dataset')
@format_docstring(**_paramdocs_unmask)
def xarray_demask_from_ds(ds, array, stackdim='_mask_stack', *, store_mask=False,
                          as_array=None, reindex=True, _upcast_bool=True):
    '''unmask (i.e., unstack) a masked (and stacked) xarray object, using mask=xarray_stored_mask(ds).
    Equivalent to xarray_unmask(array, mask=xarray_stored_mask(ds), ...)
    Also equivalent to xarray_demask_from_mask(xarray_stored_mask(ds), array, ...)
    
    ds: xarray.Dataset
        will use xarray_stored_mask(ds) as the mask.
    array: xarray.DataArray or xarray.Dataset
        object to unmask.
    stackdim: {stackdim}
    store_mask: {store_mask}
    as_array: {as_array}
    reindex: {reindex}
    _upcast_bool: {_upcast_bool}
    '''
    return xarray_demask_from_mask(xarray_stored_mask(ds), array, stackdim=stackdim, store_mask=store_mask,
                                    as_array=as_array, reindex=reindex, _upcast_bool=_upcast_bool)

@pcAccessor.register('unmask_var', totype='dataset')
@format_docstring(**_paramdocs_unmask)
def xarray_unmask_var(ds, var, mask=None, stackdim='_mask_stack', *, reindex=True, _upcast_bool=True):
    '''unmask a single variable from a Dataset. returns a DataArray of the unmasked var.
    Equivalent: xarray_unmask(ds[['_mask', var]], ...) if '_mask' in ds.data_vars
        else xarray_unmask(ds[[var]], mask=mask, ...)

    ds: xarray.Dataset
        will unmask ds[var].
    var: str
        variable whose unmasked value should be returned
    mask: None or xarray.DataArray
        mask to use. If None, use mask stored in ds['_mask'].
    stackdim: {stackdim}
    reindex: {reindex}
    _upcast_bool: {_upcast_bool}
    '''
    if var not in ds:
        raise InputConflictError(f'input dataset ds does not have variable var={var!r}.')
    to_unmask = ds[[var, '_mask']] if '_mask' in ds.data_vars else ds[[var]]
    kw_pass = dict(mask=mask, stackdim=stackdim, reindex=reindex, _upcast_bool=_upcast_bool)
    return xarray_unmask(to_unmask, **kw_pass, store_mask=False, as_array=True)

@pcAccessor.register('unmask_vars', totype='dataset')
@format_docstring(**_paramdocs_unmask)
def xarray_unmask_vars(ds, vars, mask=None, stackdim='_mask_stack', *, store_mask=False,
                       reindex=True, _upcast_bool=True):
    '''unmask multiple variables from a Dataset. returns a Dataset with unmasked vars.

    ds: xarray.Dataset
        will unmask ds[vars].
    vars: str or list of str
        variables whose unmasked values should be returned.
        if single str, treated as [vars].
    mask: None or xarray.DataArray
        mask to use. If None, use mask stored in ds['_mask'].
    store_mask: {store_mask}
    stackdim: {stackdim}
    reindex: {reindex}
    _upcast_bool: {_upcast_bool}
    '''
    if isinstance(vars, str):
        vars = [vars]
    if not all(v in ds for v in vars):
        raise InputConflictError(f'input dataset ds does not have all variables in vars={vars}.')
    to_unmask = ds[vars + (['_mask'] if '_mask' in ds.data_vars else [])]
    kw_pass = dict(mask=mask, stackdim=stackdim, reindex=reindex, _upcast_bool=_upcast_bool)
    return xarray_unmask(to_unmask, **kw_pass, store_mask=store_mask, as_array=False)



"""
File Purpose: misc tools for input/output to files.

For checking file existence or creating new files, see os_tools instead.
"""

import ast
import os
import re

from ..errors import FileContentsMissingError


''' --------------------- attempt to eval string --------------------- '''

def attempt_literal_eval(s):
    '''returns ast.literal_eval(s), or s if the literal_eval fails.'''
    try:
        return ast.literal_eval(s)
    except Exception as err:
        # failed to evaluate. Might be string, or might be int with leading 0s.
        if isinstance(err, SyntaxError):  # <-- occurs if s is an int with leading 0s.
            try:
                return int(s)
            except ValueError:
                # failed to convert to int; return original string.
                pass
        return s


''' --------------------- parse params from idl-like file --------------------- '''

def read_idl_params_file(filename, *, eval=True, as_tuples=False):
    '''parse idl file of params into a python dictionary of params.

    filename: string
        file to read from.
    eval: bool, default True
        whether to attempt to evaluate the values,
        using ast.literal_eval (safer but less flexible than eval).
        if True, try to evaluate values but use strings if evaluation fails.
        if False, values will remain as strings.
    as_tuples: bool, default False
        if True, return list of tuples instead of dictionary.
        list of tuples guaranteed to appear in the same order as in the idl file.

    File formatting notes:
        - semicolons (;) are used for comments. (idl format)
        - ignore blank lines & lines that don't assign a variable (missing '=')
        - ignores all leading & trailing whitespace in vars & values.
    '''
    filename = os.path.abspath(filename)  # <-- makes error messages more verbose, if crash later.
    # read the file lines.
    with open(filename, 'r') as file:
        lines = file.readlines()
    # remove comments
    lines = [line.split(";",1)[0] for line in lines]
    # trim whitespace
    lines = [line.strip() for line in lines]
    # remove empty lines, and remove lines without an equal sign.
    lines = [line for line in lines if line!='' and ('=' in line)]
    # split lines into vars & values
    var_value_pairs = [line.split("=",1) for line in lines]
    # cleanup whitespace in vars & values
    var_value_pairs = [(var.strip(), value.strip()) for (var, value) in var_value_pairs]

    if eval:
        var_value_pairs = [(var, attempt_literal_eval(value)) for (var, value) in var_value_pairs]

    return var_value_pairs if as_tuples else dict(var_value_pairs)

def updated_idl_params_file(filename, params, *, dst=None, missing_ok=False):
    '''create updated idl file, like at filename but updating the indicated params.
    returns updated file contents as str if dst is None, else abspath to dst.

    NEVER overwrites any existing file.

    filename: string
        file to read from.
    params: dict of params to update.
        keys are param names.
        values are param values or strings (internally, uses str(val) for each val).
        (if you need a string-valued param within the file, use pattern: '"val"' or "'val'".)
    dst: None or str
        file to write result to.
        None --> return string of updated file contents instead of creating a new file.
    missing_ok: bool
        whether it is okay for one or more of the params to be missing from the file.
        False --> if any params missing, crash.
    '''
    filename = os.path.abspath(filename)  # <-- makes error messages more verbose, if crash later.
    if (dst is not None):
        dst = os.path.abspath(dst)
        if os.path.exists(dst):
            raise FileExistsError(f"dst exists already! dst={dst!r}")
    # read the file lines.
    with open(filename, 'r') as file:
        lines = file.readlines()
    # replace param values with updated ones
    updated = set()
    for i, line in enumerate(lines):   # i is nice to track, for easy debugging later.
        match = re.fullmatch(r'(\s*)(\w+)(\s*)(=)(\s*)(.*?)(\s*;.*)?', line.rstrip('\n'))
        if match:
            s0, varname, s1, _eq_, s2, value, comment = match.groups()
            if varname in params:
                value = str(params[varname])
                lines[i] = f"{s0}{varname}{s1}={s2}{value}{comment or ''}\n"
                updated.add(varname)
    if not missing_ok:
        missing = set(params.keys()) - updated
        if missing:
            raise FileContentsMissingError(f"missing params in file: {missing!r}")
    result = ''.join(lines)   # each line already ends with '\n' by default.
    if dst is not None:
        with open(dst, 'w') as file:
            file.write(result)
        return dst
    else:
        return result


''' --------------------- parse params from python-like file --------------------- '''

def read_python_params_file(filename, *, eval=True, as_tuples=False):
    '''parse python file of params into a python dictionary of params.
    (only eval occurs via ast.literal_eval; does not evaluate arbitrary code.)

    filename: string
        file to read from.
    eval: bool, default True
        whether to attempt to evaluate the values,
        using ast.literal_eval (safer but less flexible than eval).
        if True, try to evaluate values but use strings if evaluation fails.
        if False, values will remain as strings.
    as_tuples: bool, default False
        if True, return list of tuples instead of dictionary.
        list of tuples guaranteed to appear in the same order as in the params file.

    File formatting notes:
        - hashtags (#) are used for comments. (python format)
        - ignore blank lines & lines that don't assign a variable (missing '=')
        - ignores all leading & trailing whitespace in vars & values.
    '''
    filename = os.path.abspath(filename)  # <-- makes error messages more verbose, if crash later.
    # read the file lines.
    with open(filename, 'r') as file:
        lines = file.readlines()
    # remove comments
    lines = [line.split("#",1)[0] for line in lines]
    # trim whitespace
    lines = [line.strip() for line in lines]
    # remove empty lines, and remove lines without an equal sign.
    lines = [line for line in lines if line!='' and ('=' in line)]
    # split lines into vars & values
    var_value_pairs = [line.split("=",1) for line in lines]
    # cleanup whitespace in vars & values
    var_value_pairs = [(var.strip(), value.strip()) for (var, value) in var_value_pairs]

    if eval:
        var_value_pairs = [(var, attempt_literal_eval(value)) for (var, value) in var_value_pairs]

    return var_value_pairs if as_tuples else dict(var_value_pairs)



"""
File Purpose: plotting routines by Sam
"""
import os

import numpy as np
import matplotlib.pyplot as plt
import xarray as xr

from ....dimensions import (
    DimensionFractionalIndexer,
    IONS, ELECTRON,
)
from ....plotting import PlotterManager, CMAPS, cmap
from ....tools import simple_property
from ....defaults import DEFAULTS


class EppicPlotterManagerSam(PlotterManager):
    '''eppic plotting routines by Sam'''

    # # # DEFAULTS # # #

    def _sam_timeline_defaults(ec, *, title=True):
        '''sets sam's defaults for timelines plots.
        calls plt.grid(), and plt.title(ec.title).
        '''
        plt.grid()
        if title: plt.title(ec.title)


    SAM_DEFAULTS = {
        # style for various physical parameters
        'n_style': dict(cmap='plasma', share_vlims='row'),
        'E_style': dict(cmap='cet_CET_R4', axsize=(2, 2)),
        'T_style': dict(cmap=CMAPS.get('T', 'cet_fire')),
        # ec.snap = sometimes, before doing plots of "images at a few snapshots"
        'sometimes': slice(None, None, 0.33),  # for "multifluid" case.
        'sometimes_style': dict(title_kw=dict(fontsize='small'), axsize=(1.5,1.5), subplot_title_width=15,
                                ),
        'manytimes': [0, 0.02, 0.04, 0.06, 0.08, 0.10,
                      0.15, 0.20, 0.25, 0.30, 0.35, 0.40,
                      0.50, 0.60, 0.70, 0.80, 0.90, -1],
        'manytimes_wrap': 6,
        'manytimes_style': dict(title_kw=dict(fontsize='small'), axsize=(1.5,1.5), subplot_title_width=15,
                                share_vlims='row',
                                ),
        # fft calcs for fft computations; fft style for fft plots (except "full" fft plots)
        'fft_calcs': dict(fft_keep=0.1, fft_half='x'),
        'fft_style': dict(min_n_ticks=(3,4), grid=True, robust=0.5, share_vlims='row'),
        'fft_fluids_style': dict(title_kw=dict(fontsize='xx-small'))
    }

    sam_defaults = simple_property('_sam_defaults', setdefaultvia='_default_sam_defaults',
        doc='''various defaults for sam's plotting.
        Editing ec.sam_defaults directly will change values for an instance.
        Editing ec.SAM_DEFAULTS directly will change values for all future instances.''')
    def _default_sam_defaults(self):
        '''returns a copy of self.SAM_DEFAULTS.'''
        return self.SAM_DEFAULTS.copy()

    def _extra_coords_with_title(self):
        '''return self.extra_coords with {'title': repr(self.title)} added,
        if self.title exists and 'title' not in self.extra_coords yet.
        else, just return self.extra_coords.
        '''
        extra_coords = self.extra_coords
        if 'title' in extra_coords:
            return extra_coords
        try:
            title = getattr(self, 'title')
        except AttributeError:
            return extra_coords
        else:
            return {**extra_coords, 'title': repr(title)}


    # # # DISPATCH # # #

    def sam_plots_lowres(ec, *, dst='plots_lowres/{savename}', dpi=100, bbox_inches='tight',
                        snap=slice(None, None, 0.0333), min_cost=None, max_cost=40, **kw):
        '''save sam's plots in lowres mode. return abspath to folder containing plots.'''
        log_extras = [f'Called sam_plots_lowres(min_cost={min_cost!r}, max_cost={max_cost!r})']
        with ec.using(snap=snap):
            ec.save_plots(who='sam', dst=dst, bbox_inches=bbox_inches, dpi=dpi,
                          min_cost=min_cost, max_cost=max_cost, log_extras=log_extras, **kw)
        if not os.path.isabs(dst):
            dst = os.path.join(ec.notes_dirname, dst)
        return os.path.dirname(dst)

    def sam_plots_midres(ec, *, dst='plots_midres/{savename}', dpi=200, bbox_inches='tight',
                        snap=slice(None, None, 0.01), min_cost=None, max_cost=None, **kw):
        '''save sam's plots in midres mode. return abspath to folder containing plots.'''
        log_extras = [f'Called sam_plots_midres(min_cost={min_cost!r}, max_cost={max_cost!r})']
        with ec.using(snap=snap):
            ec.save_plots(who='sam', dst=dst, bbox_inches=bbox_inches, dpi=dpi,
                          min_cost=min_cost, max_cost=max_cost, log_extras=log_extras, **kw)
        if not os.path.isabs(dst):
            dst = os.path.join(ec.notes_dirname, dst)
        return os.path.dirname(dst)

    def sam_plots_highres(ec, *, dst='plots_highres/{savename}', dpi=400, bbox_inches='tight',
                        snap=None, min_cost=None, max_cost=None, **kw):
        '''save sam's plots in highres mode. return abspath to folder containing plots.'''
        log_extras = [f'Called sam_plots_highres(min_cost={min_cost!r}, max_cost={max_cost!r})']
        with ec.using(snap=snap):
            ec.save_plots(who='sam', dst=dst, bbox_inches=bbox_inches, dpi=dpi,
                          min_cost=min_cost, max_cost=max_cost, log_extras=log_extras, **kw)
        if not os.path.isabs(dst):
            dst = os.path.join(ec.notes_dirname, dst)
        return os.path.dirname(dst)


    # # # VERY FAST PLOTS FIRST (I USED "COST=0" TO INDICATE THIS) # # #

    @known_plotter(who='sam', kinds=['timelines', 'moments', 'u', 'full'], cost=0)
    def plot_moment1(ec, **kw_plot_settings):
        '''moment1 timelines for all fluids.'''
        ec._sam_timeline_defaults()
        with ec.using(fluid=None, snap=None):
            return ec('moment1').pc.timelines(**kw_plot_settings)

    @known_plotter(who='sam', kinds=['timelines', 'moments', 'T', 'full'], cost=0)
    def plot_Ta_from_moment2(ec, **kw_plot_settings):
        '''Ta_from_moment2 timelines for all fluids.'''
        ec._sam_timeline_defaults()
        with ec.using(fluid=None, snap=None):
            return ec('Ta_from_moment2').pc.timelines(**kw_plot_settings)

    @known_plotter(who='sam', kinds=['timelines', 'moments', 'T', 'full'], cost=0)
    def plot_T_box(ec, **kw_plot_settings):
        '''T_box timelines for all fluids.'''
        ec._sam_timeline_defaults()
        with ec.using(fluid=None, snap=None):
            ec('T_from_moment2').pc.timelines(ls='--', label='(T_from_moment2)')
            ec('T_box').pc.timelines(label='(T_box)')


    # # # STATS # # #

    @known_plotter(who='sam', kinds=['timelines', 'stats', 'n'], cost=30)
    def plot_n_stats(ec, **kw_plot_settings):
        '''number density stats timelines for all fluids.'''
        with ec.using(fluid=None, snap=ec.where_loadable('n', 'current_n')):
            arr = ec('log10_stats_n')[['min', 'mean', 'median', 'max']]
        arr = arr.to_dataarray().rename('log10_stats_n')
        ec._sam_timeline_defaults()
        return arr.pc.timelines(dims=['fluid', 'variable'], **kw_plot_settings)

    @known_plotter(who='sam', kinds=['timelines', 'stats', 'E'], cost=25)
    def plot_E_stats(ec, **kw_plot_settings):
        '''E-field (vector) stats timelines.'''
        with ec.using(snap=ec.where_loadable('phi', 'current_n')):
            arr = ec('stats_E')
        arr = arr.to_dataarray().rename('stats_E')
        ec._sam_timeline_defaults()
        return arr.pc.timelines(**kw_plot_settings)

    @known_plotter(who='sam', kinds=['timelines', 'stats', 'E'], cost=27)
    def plot_E_mod_stats(ec, **kw_plot_settings):
        '''E-field (mod) stats timelines.'''
        with ec.using(snap=ec.where_loadable('phi', 'current_n')):
            arr = ec('stats_mod_E')
        arr = arr.to_dataarray().rename('stats_mod_E')
        ec._sam_timeline_defaults()
        return arr.pc.timelines(**kw_plot_settings)

    @known_plotter(who='sam', kinds=['timelines', 'stats', 'u'], cost=27)
    def plot_u_mod_stats(ec, **kw_plot_settings):
        '''stats of u (mod) timelines for all ions and electrons.
        (electrons in a subplot above ions, if mean electrons |u| >> 3 * ions |u|.)
        '''
        with ec.using(snap=ec.where_loadable('flux', 'current_n')):
            arr = ec('stats_mod_u')[['min', 'mean', 'median', 'max']]
        u_e = arr.pc.isel(fluid=ELECTRON)
        u_i = arr.pc.isel(fluid=IONS)
        if u_e['mean'].mean() > 3 * u_i['mean'].mean():  # then put electrons on their own subplot
            u_e = u_e.to_dataarray().rename('stats_mod_u')
            u_i = u_i.to_dataarray().rename('stats_mod_u')
            colors = cmap('tab10').colors
            fig, axs = plt.subplots(nrows=2, ncols=1, sharex=True,
                                    layout='constrained', squeeze=False,
                                    figsize=(10, 6), height_ratios=[1,3])
            # electrons subplot
            plt.sca(axs[0,0])
            ec._sam_timeline_defaults()
            efluid = u_e.coords['fluid'].item()
            kw = {'color': colors[0], 'cycles': [DEFAULTS.PLOT.TIMELINES_CYCLE1],
                  **kw_plot_settings}
            u_e.pc.timelines(label=f'fluid={efluid},', **kw)
            plt.xlabel('')
            # ions subplot
            plt.sca(axs[1,0])
            ec._sam_timeline_defaults(title=False)
            kw = {'cycles': [dict(color=colors[1:]), DEFAULTS.PLOT.TIMELINES_CYCLE1],
                 **kw_plot_settings}
            u_i.pc.timelines(dims=['fluid', 'variable'], **kw)
        else:  # everyone on the same plot
            arr = arr.to_dataarray().rename('stats_mod_u')
            ec._sam_timeline_defaults()
            return arr.pc.timelines(**kw_plot_settings)

    @known_plotter(who='sam', kinds=['timelines', 'stats', 'u'], cost=21)
    def plot_u_std(ec, **kw_plot_settings):
        '''std of u (vector) timelines for all ions and electrons.
        (electrons in a subplot above ions, if mean electrons std(u) >> 3 * ions std(u).)
        '''
        with ec.using(snap=ec.where_loadable('flux', 'current_n')):
            arr = ec('std_u')
        u_e = arr.pc.isel(fluid=ELECTRON)
        u_i = arr.pc.isel(fluid=IONS)
        if u_e.mean() > 3 * u_i.mean():  # then put electrons on their own subplot
            colors = cmap('tab10').colors
            fig, axs = plt.subplots(nrows=2, ncols=1, sharex=True,
                                    layout='constrained', squeeze=False,
                                    figsize=(10, 6), height_ratios=[2,3])
            # electrons subplot
            plt.sca(axs[0,0])
            ec._sam_timeline_defaults()
            efluid = u_e.coords['fluid'].item()
            kw = {'color': colors[0], 'cycles': [DEFAULTS.PLOT.TIMELINES_CYCLE1],
                  **kw_plot_settings}
            u_e.pc.timelines(label=f'fluid={efluid}', **kw)
            plt.xlabel('')
            # ions subplot
            plt.sca(axs[1,0])
            ec._sam_timeline_defaults(title=False)
            kw = {'cycles': [dict(color=colors[1:]), DEFAULTS.PLOT.TIMELINES_CYCLE1],
                 **kw_plot_settings}
            u_i.pc.timelines(**kw)
        else:  # everyone on the same plot
            ec._sam_timeline_defaults()
            return arr.pc.timelines(**kw_plot_settings)

    @known_plotter(who='sam', kinds=['timelines', 'stats', 'u'], cost=21)
    def plot_u_std_mod(ec, **kw_plot_settings):
        '''std of u (mod) timelines for all ions and electrons.
        (electrons in a subplot above ions, if mean electrons std(u) >> 3 * ions std(u).)
        '''
        with ec.using(snap=ec.where_loadable('flux', 'current_n')):
            arr = ec('std_mod_u')
        u_e = arr.pc.isel(fluid=ELECTRON)
        u_i = arr.pc.isel(fluid=IONS)
        if u_e.mean() > 3 * u_i.mean():  # then put electrons on their own subplot
            colors = cmap('tab10').colors
            fig, axs = plt.subplots(nrows=2, ncols=1, sharex=True,
                                    layout='constrained', squeeze=False,
                                    figsize=(10, 6), height_ratios=[2,3])
            # electrons subplot
            plt.sca(axs[0,0])
            ec._sam_timeline_defaults()
            efluid = u_e.coords['fluid'].item()
            u_e.pc.timelines(label=f'fluid={efluid}', color=colors[0], **kw_plot_settings)
            plt.xlabel('')
            # ions subplot
            plt.sca(axs[1,0])
            ec._sam_timeline_defaults(title=False)
            kw = {'cycles': [dict(color=colors[1:]), DEFAULTS.PLOT.TIMELINES_CYCLE1],
                 **kw_plot_settings}
            u_i.pc.timelines(**kw)
        else:  # everyone on the same plot
            ec._sam_timeline_defaults()
            return arr.pc.timelines(**kw_plot_settings)


    # # # LN_STD_DETAFRAC_N # # #

    @known_plotter(who='sam', kinds=['timelines', 'growth', 'instability', 'n'], cost=27)
    def plot_ln_std_deltafrac_n(ec, **kw_plot_settings):
        '''ln_std_deltafrac_n timelines for all fluids.'''
        with ec.using(fluid=None, snap=ec.where_loadable('n', 'current_n')):
            arr = ec('ln_std_deltafrac_n')
        ec._sam_timeline_defaults()
        return arr.pc.timelines(**kw_plot_settings)

    @known_plotter(who='sam', kinds=['timelines', 'growth', 'instability', 'n'], cost=35)
    def plot_ln_std_blur_deltafrac_n(ec, **kw_plot_settings):
        '''ln_std_blur_deltafrac_n timelines for all fluids, for blur_sigma in [0, 1, 10, 20]'''
        BLURS = [0, 1, 10, 20]
        with ec.using(fluid=None, snap=ec.where_loadable('n', 'current_n')):
            arr = xr.Dataset({blur: ec('ln_std_blur_deltafrac_n', blur_sigma=blur) for blur in BLURS})
        ec._sam_timeline_defaults()
        arr = arr.to_dataarray('blur').rename('ln_std_blur_deltafrac_n')
        return arr.pc.timelines(dims=['fluid', 'blur'], **kw_plot_settings)

    @known_plotter(who='sam', kinds=['timelines', 'growth', 'instability', 'tfbi', 'n'], cost=28)
    def plot_ddt_ln_std_deltafrac_n(ec, **kw_plot_settings):
        '''ddt_ln_std_deltafrac_n timelines for all fluids.'''
        with ec.using(fluid=None, snap=ec.where_loadable('n', 'current_n')):
            arr = ec('ddt_ln_std_deltafrac_n')
        ec._sam_timeline_defaults()
        kw_plot_settings.setdefault('robust', 5)
        return arr.pc.timelines(**kw_plot_settings)

    @known_plotter(who='sam', kinds=['timelines', 'growth', 'instability', 'tfbi', 'n'], cost=36)
    def plot_ddt_ln_std_blur_deltafrac_n(ec, **kw_plot_settings):
        '''ddt_ln_std_blur_deltafrac_n timelines for all fluids, for blur_sigma in [0, 1, 10, 20]'''
        BLURS = [0, 1, 10, 20]
        with ec.using(fluid=None, snap=ec.where_loadable('n', 'current_n')):
            arr = xr.Dataset({blur: ec('ddt_ln_std_blur_deltafrac_n', blur_sigma=blur) for blur in BLURS})
        ec._sam_timeline_defaults()
        arr = arr.to_dataarray('blur').rename('ddt_ln_std_blur_deltafrac_n')
        kw_plot_settings.setdefault('robust', 5)
        return arr.pc.timelines(dims=['fluid', 'blur'], **kw_plot_settings)


    # # # VALUES VS SPACE # # #

    @known_plotter(who='sam', kinds=['2D', 'subplots', 'movie', 'n', 'long', 'at_snap'], cost=70)
    def plot_deltafrac_n(ec, **kw_plot_settings):
        '''deltafrac_n movie for all fluids, all current snaps (in self.snap).'''
        with ec.using(fluid=None, snap=ec.where_loadable('n', 'current_n'),
                      extra_coords=ec._extra_coords_with_title()):
            arr = ec('deltafrac_n')
        kw = {'top': 0.65,  # fudge factor for prettier subplot spacing
                **ec.sam_defaults['n_style'], **kw_plot_settings}
        return arr.pc.subplots(row='fluid', **kw)

    @known_plotter(who='sam', kinds=['2D', 'subplots', 'n', 'sometimes'], cost=15)
    def plot_deltafrac_n_sometimes(ec, **kw_plot_settings):
        '''deltafrac_n at some times (ec.sam_defaults['sometimes']) for all fluids.'''
        # could just do where_loadable('n'), but we want to overlap with the other 'sometimes' plots
        somesnaps = ec.where_loadable(['n', 'phi', 'flux', 'nvsqr'])
        somesnaps = somesnaps.get(ec.sam_defaults['sometimes'])
        with ec.using(fluid=None, snap=somesnaps, extra_coords=ec._extra_coords_with_title()):
                arr = ec('deltafrac_n')
        kw = {**ec.sam_defaults['n_style'], **ec.sam_defaults['sometimes_style'],
                'title':'snap={snap}, t={t:.3g},\n{fluid}',
                **kw_plot_settings}
        return arr.pc.subplots(row='fluid', col='snap', **kw)

    @known_plotter(who='sam', kinds=['2D', 'subplots', 'n', 'manytimes'], cost=14)
    def plot_deltafrac_ne_manytimes(ec, **kw_plot_settings):
        '''deltafrac_n for electrons at "many" times (ec.sam_defaults['manytimes']).'''
        manysnaps = ec.where_loadable('n').get(DimensionFractionalIndexer(ec.sam_defaults['manytimes']))
        with ec.using(fluid='e', snap=manysnaps, extra_coords=ec._extra_coords_with_title()):
            arr = ec('deltafrac_n')
        kw = {**ec.sam_defaults['manytimes_style'],
              **ec.sam_defaults['n_style'],
              **kw_plot_settings}
        if arr.sizes['snap'] > ec.sam_defaults['manytimes_wrap']:
            kw['wrap'] = ec.sam_defaults['manytimes_wrap']
        return arr.pc.subplots(row='snap', **kw)

    @known_plotter(who='sam', kinds=['2D', 'subplots', 'E', 'sometimes'], cost=11)
    def plot_E_mod_sometimes(ec, **kw_plot_settings):
        '''mod_E at some times (ec.sam_defaults['sometimes'])'''
        # could just do where_loadable('phi'), but we want to overlap with the other 'sometimes' plots
        somesnaps = ec.where_loadable(['n', 'phi', 'flux', 'nvsqr'])
        somesnaps = somesnaps.get(ec.sam_defaults['sometimes'])
        with ec.using(snap=somesnaps, extra_coords=ec._extra_coords_with_title()):
            arr = ec('log10_mod_E')
        kw = {'top': 1.5,  # fudge factor for prettier subplot spacing
                **ec.sam_defaults['sometimes_style'], **ec.sam_defaults['E_style'], **kw_plot_settings}
        return arr.pc.subplots(col='snap', **kw)

    @known_plotter(who='sam', kinds=['2D', 'subplots', 'u', 'sometimes'], cost=16)
    def plot_u_mod_sometimes(ec, **kw_plot_settings):
        '''mod_u at some times (ec.sam_defaults['sometimes'])'''
        # could just do where_loadable('flux'), but we want to overlap with the other 'sometimes' plots
        somesnaps = ec.where_loadable(['n', 'phi', 'flux', 'nvsqr'])
        somesnaps = somesnaps.get(ec.sam_defaults['sometimes'])
        with ec.using(snap=somesnaps, extra_coords=ec._extra_coords_with_title()):
            arr = ec('mod_u')
        kw = {**ec.sam_defaults['sometimes_style'],
                'top': 1.0,  # fudge factor for prettier subplot spacing
                'title':'snap={snap}, t={t:.3g},\n{fluid}',
                **kw_plot_settings}
        return arr.pc.subplots(row='fluid', col='snap', **kw)

    @known_plotter(who='sam', kinds=['2D', 'subplots', 'T', 'sometimes'], cost=18)
    def plot_T_sometimes(ec, **kw_plot_settings):
        '''T at some times (ec.sam_defaults['sometimes'])'''
        # could just do where_loadable('nvsqr'), but we want to overlap with the other 'sometimes' plots
        somesnaps = ec.where_loadable(['n', 'phi', 'flux', 'nvsqr'])
        somesnaps = somesnaps.get(ec.sam_defaults['sometimes'])
        with ec.using(snap=somesnaps, extra_coords=ec._extra_coords_with_title()):
            arr = ec('T')
        kw = {**ec.sam_defaults['sometimes_style'],
                **ec.sam_defaults['T_style'],
                'top': 1.0,  # fudge factor for prettier subplot spacing
                'title':'snap={snap}, t={t:.3g},\n{fluid}',
                **kw_plot_settings}
        return arr.pc.subplots(row='fluid', col='snap', **kw)


    # # # FFT (deltafrac_n) # # #

    @known_plotter(who='sam', kinds=['2D', 'subplots', 'fft', 'n', 'sometimes'], cost=22)
    def plot_deltafrac_n_abs_radfft_sometimes(ec, **kw_plot_settings):
        '''abs_radfft_deltafrac_n at some times (ec.sam_defaults['sometimes']) for all fluids.
        Not the full region of k-space; see also plot_deltafrac_n_abs_radfft_full.
        '''
        # could just do where_loadable('n'), but we want to overlap with the other 'sometimes' plots
        somesnaps = ec.where_loadable(['n', 'phi', 'flux', 'nvsqr'])
        somesnaps = somesnaps.get(ec.sam_defaults['sometimes'])
        with ec.using(fluid=None, snap=somesnaps, extra_coords=ec._extra_coords_with_title()):
            arr = ec('abs_radfft_deltafrac_n', **ec.sam_defaults['fft_calcs'])
        kw = {**ec.sam_defaults['sometimes_style'],
              **ec.sam_defaults['fft_style'],
              **ec.sam_defaults['fft_fluids_style'],
              'title':'snap={snap},\nt={t:.3g},\n{fluid}',
              **kw_plot_settings}
        result = arr.pc.subplots(row='fluid', col='snap', **kw)
        result.scatter_max() # put markers at maximum on each subplot.
        return result

    @known_plotter(who='sam', kinds=['2D', 'subplots', 'fft', 'n', 'sometimes'], cost=27)
    def plot_deltafrac_n_blurk_abs_radfft_sometimes(ec, **kw_plot_settings):
        '''blurk_abs_radfft_deltafrac_n at some times (ec.sam_defaults['sometimes']) for all fluids.'''
        blur = {'blur_sigma': 1}
        # could just do where_loadable('n'), but we want to overlap with the other 'sometimes' plots
        somesnaps = ec.where_loadable(['n', 'phi', 'flux', 'nvsqr'])
        somesnaps = somesnaps.get(ec.sam_defaults['sometimes'])
        with ec.using(fluid=None, snap=somesnaps, extra_coords=ec._extra_coords_with_title()):
            arr = ec('blurk_abs_radfft_deltafrac_n', **ec.sam_defaults['fft_calcs'], **blur)
        kw = {**ec.sam_defaults['sometimes_style'],
              **ec.sam_defaults['fft_style'],
              **ec.sam_defaults['fft_fluids_style'],
              'title':'snap={snap},\nt={t:.3g},\n{fluid}',
              **kw_plot_settings}
        result = arr.pc.subplots(row='fluid', col='snap', **kw)
        result.scatter_max() # put markers at maximum on each subplot.
        return result

    @known_plotter(who='sam', kinds=['2D', 'subplots', 'fft', 'n', 'sometimes'], cost=23)
    def plot_deltafrac_n_abs_radfft_full_sometimes(ec, **kw_plot_settings):
        '''abs_radfft_deltafrac_n at some times (ec.sam_defaults['sometimes']) for all fluids.
        Includes the full region of k-space; see also plot_deltafrac_n_abs_radfft_full.
        '''
        fft_full = dict(fft_slices={})  # <-- full region of k-space
        # could just do where_loadable('n'), but we want to overlap with the other 'sometimes' plots
        somesnaps = ec.where_loadable(['n', 'phi', 'flux', 'nvsqr'])
        somesnaps = somesnaps.get(ec.sam_defaults['sometimes'])
        with ec.using(fluid=None, snap=somesnaps, extra_coords=ec._extra_coords_with_title()):
            arr = ec('abs_radfft_deltafrac_n', **fft_full)
        kw = {**ec.sam_defaults['sometimes_style'],
                'top': 1.0,  # fudge factor for prettier subplot spacing
                'title':'snap={snap}, t={t:.3g},\n{fluid}',
                **kw_plot_settings}
        return arr.pc.subplots(row='fluid', col='snap', **kw)

    @known_plotter(who='sam', kinds=['2D', 'subplots', 'fft', 'n', 'movie', 'long'], cost=71)
    def plot_deltafrac_n_abs_radfft(ec, **kw_plot_settings):
        '''abs_radfft_deltafrac_n movie for all fluids, all current snaps (in self.snap).
        Not the full region of k-space.
        '''
        with ec.using(fluid=None, snap=ec.where_loadable('n', 'current_n'),
                      extra_coords=ec._extra_coords_with_title()):
            arr = ec('abs_radfft_deltafrac_n', **ec.sam_defaults['fft_calcs'])
        kw = {**ec.sam_defaults['fft_style'],
              **ec.sam_defaults['fft_fluids_style'],
              'title': '{fluid}',
              'top': 0.5, # fudge factor for prettier subplot spacing
              **kw_plot_settings}
        return arr.pc.subplots(row='fluid', **kw)

    @known_plotter(who='sam', kinds=['2D', 'subplots', 'fft', 'n', 'movie'], cost=75)
    def plot_deltafrac_n_blurk_abs_radfft(ec, **kw_plot_settings):
        '''blurk_abs_radfft_deltafrac_n movie for all fluids, all current snaps (in self.snap).
        Not the full region of k-space. Uses blur_sigma=1.
        '''
        blur = {'blur_sigma': 1}
        with ec.using(fluid=None, snap=ec.where_loadable('n', ec.snap_list()),
                      extra_coords=ec._extra_coords_with_title()):
            arr = ec('blurk_abs_radfft_deltafrac_n', **ec.sam_defaults['fft_calcs'], **blur)
        kw = {**ec.sam_defaults['fft_style'],
              **ec.sam_defaults['fft_fluids_style'],
              'title': '{fluid}',
              'top': 0.5, # fudge factor for prettier subplot spacing
              **kw_plot_settings}
        return arr.pc.subplots(row='fluid', **kw)


    # # # FFT (|E|**2) # # #

    @known_plotter(who='sam', kinds=['2D', 'subplots', 'fft', 'E', 'manytimes'], cost=21)
    def plot_E_mod2_abs_radfft_manytimes(ec, **kw_plot_settings):
        '''abs_radfft_delta_mod2_E at "many" times (ec.sam_defaults['manytimes']).'''
        manysnaps = ec.where_loadable('phi').get(DimensionFractionalIndexer(ec.sam_defaults['manytimes']))
        with ec.using(snap=manysnaps, extra_coords=ec._extra_coords_with_title()):
            arr = ec('abs_radfft_delta_mod2_E', **ec.sam_defaults['fft_calcs'])
        kw = {**ec.sam_defaults['manytimes_style'],
              **ec.sam_defaults['fft_style'],
              'title':'snap={snap},\nt={t:.3g}',
              **kw_plot_settings}
        if arr.sizes['snap'] > ec.sam_defaults['manytimes_wrap']:
            kw['wrap'] = ec.sam_defaults['manytimes_wrap']
        result = arr.pc.subplots(row='snap', **kw)
        result.scatter_max() # put markers at maximum on each subplot.
        return result

    @known_plotter(who='sam', kinds=['2D', 'subplots', 'fft', 'E', 'manytimes'], cost=12)
    def plot_E_mod2_blurk_abs_radfft_manytimes(ec, **kw_plot_settings):
        '''abs_radfft_delta_mod2_E at "many" times (ec.sam_defaults['manytimes']).'''
        manysnaps = ec.where_loadable('phi').get(DimensionFractionalIndexer(ec.sam_defaults['manytimes']))
        blur = {'blur_sigma': 1}
        with ec.using(snap=manysnaps, extra_coords=ec._extra_coords_with_title()):
            arr = ec('blurk_abs_radfft_delta_mod2_E', **ec.sam_defaults['fft_calcs'], **blur)
        kw = {**ec.sam_defaults['manytimes_style'],
              **ec.sam_defaults['fft_style'],
              'title':'snap={snap},\nt={t:.3g}',
              **kw_plot_settings}
        if arr.sizes['snap'] > ec.sam_defaults['manytimes_wrap']:
            kw['wrap'] = ec.sam_defaults['manytimes_wrap']
        result = arr.pc.subplots(row='snap', **kw)
        result.scatter_max() # put markers at maximum on each subplot.
        return result



"""
File Purpose: reload, ImportFailed
"""

import importlib
import inspect
import sys
import warnings

from ..errors import ImportFailedError, InputMissingError
from ..defaults import DEFAULTS


''' --------------------- reloading --------------------- '''

def enable_reload(package='PlasmaCalcs'):
    '''smashes the import cache for the provided package.
    All modules starting with this name will be removed from sys.modules.
    This does not actually reload any modules.
    However, it means the next time import is called for those modules, they will be reloaded.
    returns tuple listing of all names of affected modules.

    package: str or module
        the package for which to enable reload. if module, use package.__name__.
    '''
    if inspect.ismodule(package):
        package = package.__name__
    l = tuple(key for key in sys.modules.keys() if key.startswith(package))
    for key in l:
        del sys.modules[key]
    return l

def reload(package='PlasmaCalcs', *, return_affected=False, maintain_defaults=True):
    '''reloads the provided package.
    Similar to enable_reload(package); import package.

    maintain_defaults: bool, default True
        if True, update DEFAULTS in the reloaded package with any DEFAULTS set before reloading.
        Does nothing if the original package does not have a 'DEFAULTS' attribute.
        e.g. pc.reload() stores defaults0=pc.DEFAULTS, then reloads, then pc.DEFAULTS.update(defaults0).

    returns the reloaded package.
    if return_affected also return a list of all affected package names.

    NOTE: this reloads package but doesn't do it "in-place" (i.e. doesn't change the package variable).
    For example:
        import PlasmaCalcs as pc
        import mypackage as myp0
        myp1 = pc.reload(myp0)
        myp1 is myp0 --> False, because myp0 points to the pre-reload version of the package.

    Thus, to use this method, you should instead assign the package to the result, for example:
        import PlasmaCalcs as pc
        pc = pc.reload()

        # or, to reload a different package
        import mypackage as myp
        myp = pc.reload(myp)

    package: str or module
        the package to reload. if module, use package.__name__.
    '''
    if inspect.ismodule(package):
        package = package.__name__
    module = importlib.import_module(package)
    defaults0 = module.DEFAULTS if (maintain_defaults and 'DEFAULTS' in module.__dict__) else None
    affected = enable_reload(package)
    result = importlib.import_module(package)
    if defaults0 is not None:
        result.DEFAULTS.update(defaults0)
    return (result, affected) if return_affected else result


''' --------------------- relative loading inside package --------------------- '''

def import_relative(name, globals):
    '''import a module relative to the caller's package; caller must provide globals().
    Examples: inside SymSolver.tools.arrays,
        import_relative('.numbers', globals()) <- equivalent -> import SymSolver.tools.numbers
        import_relative('..errors', globals()) <- equivalent -> import SymSolver.errors
    '''
    package = globals['__package__']
    return importlib.import_module(name, package=package)


''' --------------------- import failure handling --------------------- '''

class ImportFailed():
    '''set modules which fail to import to be instances of this class;
    initialize with modulename, additional_error_message.
    when attempting to call or access any attribute of the ImportFailed object,
        raises ImportFailedError('. '.join(modulename, additional_error_message)).
    Also, if DEFAULTS.IMPORT_FAILURE_WARNINGS, make warning immediately when initialized.

    err: None, Exception, or str
        if provided, include str(err) when raising ImportFailedError
    locals: None or dict
        if provided, set locals['__pdoc__'][modulename] = False
        if abbrv provided, use abbrv instead of modulename.
        Use this to tell pdoc to skip the ImportFailed instance when creating docs pages.
    abbrv: None or str
        if provided, set locals['__pdoc__'][abbrv] = False

    Example:
        try:
            import h5netcdf
        except ImportError as err:
            h5netcdf = ImportFailed('h5netcdf', 'This module is required for compressing data.', err=err)

        h5netcdf.load(...)   # << attempt to use h5netcdf
        # if h5netcdf was imported successfully, it will work fine.
        # if h5netcdf failed to import, this error will be raised:
            ImportFailedError: h5netcdf. This module is required for compressing data.
            The original ImportError error was: No module named 'h5netcdf'
    '''
    def __init__(self, modulename, additional_error_message='', *, err=None, locals=None, abbrv=None):
        self.modulename = modulename
        self.additional_error_message = additional_error_message
        self.err = err
        if DEFAULTS.IMPORT_FAILURE_WARNINGS:
            warnings.warn(f'Failed to import module {self.error_message()}')
        # handle __pdoc__
        if (abbrv is not None) and (locals is None):
            raise InputMissingError('if providing abbrv, must also provide locals.')
        if locals is not None:
            key = modulename if abbrv is None else abbrv
            locals.setdefault('__pdoc__', {})[key] = False

    def error_message(self):
        str_add = str(self.additional_error_message)
        if len(str_add) > 0:
            str_add = '. ' + str_add
        result = self.modulename + str_add
        if self.err is not None:
            result += f'\nThe original ImportError was: {self.err}'
        return result

    def __getattr__(self, attr):
        '''tells how to do self.attr when it would otherwise fail.
        Here, raise ImportFailedError.
        '''
        raise ImportFailedError(self.error_message())

    def __call__(self, *args__None, **kw__None):
        '''tells how to call self, e.g. self(...). Here, raise ImportFailedError.'''
        raise ImportFailedError(self.error_message())

    def __repr__(self):
        return f'{type(self).__name__}({repr(self.modulename)})'



"""
File Purpose: FluidsLoader
"""

from ..quantity_loader import QuantityLoader

from ...dimensions import ELECTRON, IONS
from ...tools import simple_property, xarray_sum

class FluidsLoader(QuantityLoader):
    '''load fluid-related quantities, e.g. fluids_sum_{var}'''

    @known_pattern(r'sum_(s|fluid)_(.+)', deps=[1], reduces_dims=['fluid'])  # e.g. sum_s_n, sum_fluid_T
    def get_sum_fluid_var(self, var, *, _match=None):
        '''var summed across self.fluid. Equivalent: self(var).pc.sum('fluid').
        (if not self.fluid_is_iterable(), result numerically equivalent to self(var).)
        aliases: sum_fluid_var or sum_s_var.
        '''
        _s_or_fluid, var, = _match.groups()
        return xarray_sum(self(var), dim='fluid')
    
    @known_pattern(r'sum_fluids_(.+)', deps=[0], ignores_dims=['fluid'])  # e.g. sum_fluids_n
    def get_sum_fluids_var(self, var, *, _match=None):
        '''var summed across self.fluids. Equivalent: self(var, fluid=None).pc.sum('fluid').'''
        var, = _match.groups()
        return xarray_sum(self(var), dim='fluid')

    @known_pattern(r'sum_(j|jfluid)_(.+)', deps=[1], reduces_dims=['jfluid'])  # e.g. sum_j_n, sum_jfluid_T
    def get_sum_jfluid_var(self, var, *, _match=None):
        '''var summed across self.jfluid. Equivalent: self(var).pc.sum('jfluid').
        (if not self.jfluid_is_iterable(), result numerically equivalent to self(var).)
        aliases: sum_jfluid_var or sum_j_var.
        '''
        _j_or_jfluid, var, = _match.groups()
        return xarray_sum(self(var), dim='jfluid')

    @known_pattern(r'sum_jfluids_(.+)', deps=[0], ignores_dims=['jfluid'])  # e.g. sum_jfluids_n
    def get_sum_jfluids_var(self, var, *, _match=None):
        '''var summed across self.jfluids. Equivalent: self(var, jfluid=None).pc.sum('jfluid').'''
        var, = _match.groups()
        return xarray_sum(self(var), dim='jfluid')

    @known_pattern(r'sum_ions_(.+)', deps=[0], ignores_dims=['fluid'])  # e.g. sum_ions_n
    def get_sum_ions_var(self, var, *, _match=None):
        '''var summed across all ions in self.fluids.
        Equivalent: self(var, fluid=self.fluids.ions()).pc.sum('fluid').
        '''
        var, = _match.groups()
        return xarray_sum(self(var, fluid=self.fluids.ions()), dim='fluid')

    @known_pattern(r'sum_neutrals_(.+)', deps=[0], ignores_dims=['fluid'])  # e.g. sum_neutrals_n
    def get_sum_neutrals_var(self, var, *, _match=None):
        '''var summed across all neutrals in self.fluids.
        Equivalent: self(var, fluid=self.fluids.neutrals()).pc.sum('fluid').
        '''
        var, = _match.groups()
        return xarray_sum(self(var, fluid=self.fluids.neutrals()), dim='fluid')

    # # # nnefrac # # #
    @known_var(deps=['n', ('n', {'fluid': ELECTRON})])
    def get_nnefrac(self):
        '''ratio of number density to electron number density: n / ne.'''
        return self('n') / self('n', fluid=ELECTRON)

    @known_var(deps=[('nnefrac', {'fluid': IONS})], ignores_dims=['fluid'])
    def get_ionnefrac(self):
        '''ratio of n_ions / ne. Equivalent: self('nnefrac', fluid=self.fluids.ions()).'''
        return self('nnefrac', fluid=IONS)

    cls_behavior_attrs.register('nnefrac_tiny_thresh', default=0.01)
    nnefrac_tiny_thresh = simple_property('_nnefrac_tiny_thresh', default=0.01,
        doc='''threshhold for where_nnefrac_tiny and where_nnefrac_significant.
        tiny if n/ne < nnefrac_tiny_thresh, significant if n/ne >= nnefrac_tiny_thresh.''')

    @known_var(deps=['nnefrac'])
    def get_nnefrac_tiny(self):
        '''boolean array telling whether n/ne < self.nnefrac_tiny_thresh'''
        return self('nnefrac') < self.nnefrac_tiny_thresh

    @known_var(deps=['ionnefrac'])
    def get_ionnefrac_tiny(self):
        '''boolean array telling whether n_ions/ne < self.nnefrac_tiny_thresh'''
        return self('ionnefrac') < self.nnefrac_tiny_thresh

    @known_var(deps=['nnefrac_tiny'], aliases=['nnefrac_not_tiny'])
    def get_nnefrac_significant(self):
        '''boolean array telling whether n/ne >= self.nnefrac_tiny_thresh.
        Equivalent: self('not_nnefrac_tiny')
        '''
        return self('not_nnefrac_tiny')

    @known_var(deps=['ionnefrac_tiny'], aliases=['ionnefrac_not_tiny'])
    def get_ionnefrac_significant(self):
        '''boolean array telling whether n_ions/ne >= self.nnefrac_tiny_thresh.
        Equivalent: self('not_ionnefrac_tiny')
        '''
        return self('not_ionnefrac_tiny')



"""
File Purpose: base quantities for multifluid analysis of single-fluid mhd
"""

import xarray as xr

from ..mhd_bases import MhdBasesLoader
from ...dimensions import SINGLE_FLUID
from ...errors import (
    InputError, InputMissingError, FormulaMissingError,
)
from ...tools import (
    simple_property,
    UNSET,
)

class MhdMultifluidBasesLoader(MhdBasesLoader):
    '''base quantities for multifluid analysis of single-fluid mhd.
    if self.fluid is SINGLE_FLUID, results are equivalent to single-fluid mhd results.
    '''
    @known_var(load_across_dims=['fluid'])
    def get_q(self):
        '''charge of fluid particle. fluid.q [converted to self.units] if it exists.
        if fluid.q is None, give nan. If fluid.q does not exist, crash.
        '''
        f = self.fluid
        if hasattr(f, 'q'):
            q = f.q if f.q is None else (f.q * self.u('qe'))
            return xr.DataArray(q, attrs=self.units_meta())
        else:
            raise FormulaMissingError(f'charge q for fluid {f} without fluid.q attribute.')

    # # # BOOKKEEPING FOR VALUES WHICH ASSUME SINGLE FLUID # # #
    @known_pattern(r'(SF|SINGLE_FLUID)_(.+)',  deps=[(1, {'fluid': SINGLE_FLUID})], ignores_dims=['fluid'])
    def get_single_fluid_var(self, var, *, _match=None, **kw):
        '''SF_{var} (or SINGLE_FLUID_{var}) --> {var} using self.fluid=SINGLE_FLUID.
        (any kwargs get passed to self(var, **kw))
        '''
        _sf, here, = _match.groups()
        with self.using(fluid=SINGLE_FLUID):
            result = self(here, **kw)
        return result

    # # # NEUTRALS # # #
    @known_var(deps=['m'], ignores_dims=['fluid'], aliases=['m_n'])
    def get_m_neutral(self):
        '''mass, of a "single neutral particle". Equivalent to self('m') for neutral fluid.
        Only works if self.fluids (or self.jfluids) contains exactly 1 neutral Specie.
        '''
        return self.get_neutral('m')

    @known_var(deps=['n'], ignores_dims=['fluid'], aliases=['n_n'])
    def get_n_neutral(self):
        '''number density of neutral fluid. Equivalent to self('n') for neutral fluid.
        Only works if self.fluids (or self.jfluids) contains exactly 1 neutral Specie.
        '''
        return self.get_neutral('n')

    cls_behavior_attrs.register('assume_un', default=None)
    assume_un = simple_property('_assume_un', default=None,
        doc='''None, 'u', or xarray.DataArray to assume for u_neutral.
        value to assume for u_neutral (used when calculating u_neutral, and E_un0 if E_un0_mode=None).
        None --> cannot determine u_n (crash if trying to get it).
        'u' --> assume u_n = self('u'). Maybe reasonable for weakly-ionized plasma.
                in this case, E_un0 = E_u0, i.e. E(u=0 frame) == E(u_n=0 frame).
        xarray.DataArray --> assume these values for u_n.
                Should have 'x', 'y', 'z' components.''')

    @known_var(aliases=['u_n'])
    def get_u_neutral(self):
        '''velocity of neutral fluid. Depends on self.assume_un:
            None --> cannot get value; raise FormulaMissingError.
            'u' --> assume u_n == self('SINGLE_FLUID_u')
            else --> return self.assume_un
        '''
        assume_un = self.assume_un
        if assume_un is None:
            raise FormulaMissingError('u_neutral, when self.assume_un not provided (=None).')
        elif assume_un == 'u':
            return self('SINGLE_FLUID_u')
        else:
            return assume_un  # [TODO] handle assume_un components != self.component



"""
Package Purpose: other PlasmaCalculators.
E.g. allow user to input data directly during __init__.

These can require addons, but should not require hookups.
    hookups are for reading specific kinds of data.
    other_calculators are for generic data.
E.g., a hookup might rely on one of the other_calculators.
"""
from .from_dataset import (
    DimensionlessFromDatasetCalculator, ComponentHavingFromDatasetCalculator,
    FromDatasetCalculator, MultifluidFromDatasetCalculator,
    VectorlessFromDatasetCalculator, VectorlessMultifluidFromDatasetCalculator,
)
from .instability_calculator import (
    VectorlessInstabilityCalculator, InstabilityCalculator,
)



"""
File Purpose: loading quantities with masks
"""
import xarray as xr

from .quantity_loader import QuantityLoader
from ..errors import InputError
from ..tools import (
    simple_property,
    UNSET,
    xarray_mask, xarray_unmask, xarray_has_mask, xarray_store_mask,
)

class MaskLoader(QuantityLoader):
    '''manage masks for quantities.'''

    # # # PROPERTIES # # #
    cls_behavior_attrs.register('mask', default=None)
    cls_behavior_attrs.register('masking', default='stacked')
    cls_behavior_attrs.register('output_mask', default=False)

    @property
    def _extra_kw_for_quantity_loader_call(self):
        '''extra kwargs which can be used to set attrs self during self.__call__.
        The implementation here returns ['assert_masking'] + any values from super().
        '''
        return ['assert_masking'] + super()._extra_kw_for_quantity_loader_call

    mask = simple_property('_mask', default=None,
        doc='''None or xarray.DataArray of bools,
        if self.masking, apply mask to results (with mask dims) at top-level (call_depth=1)
            (internal calls still do full calculations, so derivatives will still work.)
        applying mask means calling self.apply_mask(result);
            masking = True or 'stacked' --> xarray_mask(result, stack=True); i.e., drops all masked points.
                        (and, result will have '_mask_stack' dimension instead of mask.dims dimensions)
            masking = 'simple' --> xarray_mask(result, stack=False); i.e., masked points just replaced by nan.
        never applies mask if self.mask = None.
        setting self.mask = value is equivalent to calling self.set_mask(value).''')
    @mask.setter
    def mask(self, value):
        '''call self.set_mask(value). Probably just sets self._mask = value.'''
        self.set_mask(value)

    masking = simple_property('_masking', default='stacked',
        doc='''how to apply self.mask to results at top level, if mask exists.
        True --> alias for 'stacked'. This is the default.
        'stacked' --> apply stacked mask to self.stack(result), dropping masked points.
        'simple' --> apply mask to result, filling masked regions with np.nan.
        False --> do not apply mask, even if self.mask exists.''')
    @masking.setter
    def masking(self, value):
        '''set self._masking = value. If value is True, set to 'stacked'.'''
        if value == True:
            value = 'stacked'
        if value not in ('stacked', 'simple', False):
            raise InputError(f'self.masking must be True, False, "stacked", or "simple", not {value!r}')
        self._masking = value

    assert_masking = simple_property('_assert_masking', default=False,
        doc='''whether to assert self.masking != False and mask is not None, when getting values.''')

    output_mask = simple_property('_output_mask', default=False,
        doc='''whether to store_mask during xarray_mask calls if self.masking.
        False --> never store mask in results
        True --> always store mask in top-level results
                (all results will be xarray.Dataset objects with '_mask' data_var)
        None --> only store mask in results which would have been Datasets anyways.''')

    def set_mask(self, mask):
        '''sets self.mask = mask, and increments self._mask_cache_state (if checks succeed).
        Also does some checks:
            - mask is None or xarray.DataArray. If None, don't do any other checks.
        Also may alter mask slightly:
            - discard non-dimension coords
        '''
        if mask is None:
            self._mask = None
        else:
            if not isinstance(mask, xr.DataArray):
                raise InputError(f'mask must be None or xarray.DataArray, not {type(mask)}')
            nondim_coords = tuple(c for c in mask.coords if mask.coords[c].ndim == 0)
            if nondim_coords:
                mask = mask.drop_vars(nondim_coords)
            self._mask = mask
        self._mask_cache_state = 1 + getattr(self, '_mask_cache_state', -1)

    # # # STACKING AND MASKING -- HOOKUP TO __call__ # # #
    def _call_postprocess_toplevel(self, result, *, var, name=UNSET, item=UNSET):
        '''additional postprocessing for self.__call__ when call_depth=1.
        called from self._call_postprocess, after doing other postprocessing, when call_depth=1.

        result: any value, probably an xarray.DataArray
            result from self.__call__, after other postprocessing.
        var, name, item: UNSET or value
            passed directly from self.__call__.
            Don't need to handle these here because self._call_postprocess will handle it.

        The implementation here applies masking, if applicable.
            see help(type(self).mask) 
        '''
        if self.assert_masking:
            if self.masking == False:
                raise ValueError('assert_masking=True, but self.masking=False')
            if self.mask is None:
                raise ValueError('assert_masking=True, but self.mask=None')
        if self.mask is not None and self.masking != False:
            result_dims = getattr(result, 'dims', [])
            if all(d in result_dims for d in self.mask.dims):
                result = self.apply_mask(result)
        return super()._call_postprocess_toplevel(result, var=var, name=name, item=item)

    # # # STACKING AND MASKING -- METHODS # # #
    def apply_mask(self, arr, masking=UNSET):
        '''apply self.mask to arr, using self.masking to determine masking.
        arr must have dims from self.mask.dims.
        masking: UNSET, bool, 'stacked', or 'simple'
            UNSET --> use self.masking.
            True --> alias for 'stacked'
            'stacked' --> apply stacked mask to self.stack(arr), dropping masked points.
            'simple' --> apply mask to arr, filling masked regions with np.nan.
        will crash with ValueError if masking corresponds to False, or if self.mask is None.
        '''
        if masking is UNSET:
            masking = self.masking
        else:
            with self.using(masking=masking):
                masking = self.masking   # handles aliasing for True appropriately.
        if masking == False:
            raise ValueError('apply_mask when masking=False')
        if self.mask is None:
            raise ValueError('apply_mask when self.mask=None')
        output_mask = self.output_mask
        if masking == 'stacked':
            return xarray_mask(arr, self.mask, stack=True, store_mask=output_mask)
        elif masking == 'simple':
            return xarray_mask(arr, self.mask, stack=False, store_mask=output_mask)
        else:
            assert False, f'coding error... invalid masking should be caught by property setter.'

    def unmask(self, arr, **kw_xarray_unmask):
        '''restore arr to the same shape as unmasked unstacked arrays. masked vals will still be np.nan.
        Equivalent: xarray_unmask(arr) if arr has '_mask' data_var, else xarray_unmask(arr, mask=self.mask)
        '''
        if xarray_has_mask(arr):
            return xarray_unmask(arr, **kw_xarray_unmask)
        else:
            return xarray_unmask(arr, mask=self.mask, **kw_xarray_unmask)

    def store_mask(self, arr):
        '''equivalent: xarray_store_mask(arr, self.mask). Also equivalent: arr.pc.store_mask(self.mask)'''
        return xarray_store_mask(arr, self.mask)

    # # # KNOWN PATTERS # # #
    @known_pattern(r'(simple_|stacked_)?mask_(.+)', deps=[1])  # mask_{var} or simple_mask_{var} or stacked_mask_{var}
    def get_mask_var(self, var, *, _match=None):
        '''mask_var --> masked version of self(var).
        mask_var and stacked_mask_var = equivalent to self(var, masking='stacked').
        simple_mask_var = equivalent to self(var, masking='simple').
        '''
        mode, here = _match.groups()
        if mode is None:
            mode = 'stacked'
        return self(here, masking=mode)



"""
File Purpose: fft of numpy arrays.
"""

import numpy as np

from ..iterables import is_iterable
from ..math import float_rounding as math_float_rounding
from ...errors import InputConflictError, InputMissingError


''' --------------------- fft --------------------- '''

def fftN(array, ds, axes=None, *, rad=False, **kw_np_fftn):
    '''calculates fft along N dimensions, as well as the corresponding frequencies.

    array: array-like
        take the fft of this array (along the indicated axes).
    ds: list-like with length == len(axes) (or array.ndim if axes is None)
        specifies the spacing between elements of array (in position-space), along each axis.
        This is necessary in order to provide the frequency coordinates for the result.
    axes: None, or iterable of ints
        axes to take fft over. (Negative values are supported & count from end.)
        None --> take fft over all axes.
    rad: bool
        if True, convert frequencies to "radians" by multiplying them by 2 * pi.
        E.g., for fft in space, rad=False gives 1/wavelength; rad=True gives wavenumber k.
    
    additional kwargs passed to np.fft.fftn.

    returns (shifted fftn(array, axes=axes, **kw_np_fftn), (freq0, freq1, ..., freqN)),
        where freq0, freq1, ..., freqN are the frequencies corresponding to the N axes,
        and all freqs & values are shifted with np.fft.fftshift to put the 0-frequency in the center.
    '''
    array = np.asanyarray(array)
    if axes is None:
        axes = list(range(array.ndim))
    if len(ds) != len(axes):
        raise InputConflictError(f"len(ds)={len(ds)} != len(axes)={len(axes)}")
    Ns = [array.shape[ax] for ax in axes]
    shifted_freqs = tuple(fftfreq_shifted(N, d, rad=rad) for N, d in zip(Ns, ds))
    # take fft & shift appropriately.
    vals = np.fft.fftn(array, axes=axes, **kw_np_fftn)
    vals = np.fft.fftshift(vals, axes=axes)
    return vals, shifted_freqs

def fft2(array, dx, dy, axes=(-2,-1), *, rad=False, **kw_np_fftn):
    '''calculates fft along 2 dimensions, as well as the corresponding frequencies.
    Equivalent to:
        fftN(array, (dx, dy), axes=axes, rad=rad, **kw_np_fftn).

    array: array-like
        take the fft of this array.
    dx, dy: numbers
        spacing between elements of array (in position-space), along each axis.
        This is necessary in order to provide the frequency coordinates for the result.
    axes: iterable of ints, default (-2,-1)
        axes to take fft over. (Negative values are supported & count from end.)
    rad: bool
        if True, convert frequencies to "radians" by multiplying them by 2 * pi.
        E.g., for fft in space, rad=False gives 1/wavelength; rad=True gives wavenumber k.
    
    additional kwargs passed to np.fft.fftn.

    returns (shifted fft2(array, axes=axes, **kw_np_fftn), (freqx, freqy)),
        where freqx, freqy are the frequencies corresponding to the x and y axes,
        and all freqs & values are shifted with np.fft.fftshift to put the 0-frequency in the center.
    '''
    return fftN(array, (dx, dy), axes=axes, rad=rad, **kw_np_fftn)

def fft1(array, dx, axis=-1, *, rad=False, **kw_np_fftn):
    '''calculates fft along 1 dimension, as well as the corresponding frequencies.
    Equivalent to:
        vals, freqs = fftN(array, (dx,), axes=(axis,), rad=rad, **kw_np_fftn);
        return vals, freqs[0]

    array: array-like
        take the fft of this array.
    dx: number
        spacing between elements of array (in position-space), along the axis.
        This is necessary in order to provide the frequency coordinates for the result.
    axis: int, default -1
        axis to take fft over. (Negative values are supported & count from end.)
    rad: bool
        if True, convert frequencies to "radians" by multiplying them by 2 * pi.
        E.g., for fft in space, rad=False gives 1/wavelength; rad=True gives wavenumber k.
    
    additional kwargs passed to np.fft.fftn.

    returns (shifted fft1(array, axis=axis, **kw_np_fftn), freqx),
        where freqx is the frequencies corresponding to the x axis,
        and all freqs & values are shifted with np.fft.fftshift to put the 0-frequency in the center.
    '''
    vals, freqs = fftN(array, (dx,), axes=(axis,), rad=rad, **kw_np_fftn)
    return vals, freqs[0]

def fftfreq_shifted(N, d, *, rad=False):
    '''returns the shifted frequencies corresponding to the fft of an array of length N,
    with spacing d between elements.
    if rad: convert frequencies to "radians" by multiplying them by 2 * pi.

    "shifted" means that the 0-frequency is in the center of the result.
    '''
    freq = np.fft.fftfreq(N, d)
    if rad:
        freq *= 2 * np.pi
    return np.fft.fftshift(freq)


''' --------------------- inverse fft --------------------- '''

def ifftN(array, df=None, axes=None, *, rad=False, x0=0, ds=None, **kw_np_ifftn):
    '''calculates inverse fft along N dimensions, as well as the corresponding positions.

    Caution: ifftN(fftN(arr)) == arr only approximately, due to floating point rounding errors.
        Can at least ensure coordinate alignment by providing ds during ifftN(fftN(arr), ds=...)

    array: array-like
        take the inverse fft of this array (along the indicated axes).
        the 0-frequency component should be centered in each axis,
            as per the output of np.fft.fftshift(np.fft.fft(position_space_array)).
    df: None, number, or list-like with length == len(axes)
        specifies the spacing between elements of array (in frequency-space), along each axis.
        None --> use ds instead. (Must provide ds or df.)
    axes: None, or iterable of ints
        axes to take inverse fft over. (Negative values are supported & count from end.)
        None --> take inverse fft over all axes.
    rad: bool
        if True, interpret frequency-spacing (df) like it is "in radians",
            dividing it by 2 * pi before converting to position-space.
    x0: None, number, or list-like with length == len(axes)
        if provided, alter position-space coordinates by adding a constant offset,
            such that the 0'th position for axes[i] equals x0[i].
    ds: None, number, or list-like with length == len(axes)
        specifies the spacing between elements of result (in position-space), along each axis.
        None --> use df instead. (Must provide ds or df.)
    
    additional kwargs passed to np.fft.ifftn.

    returns (shifted ifftn(array, axes=axes, **kw_np_ifftn), (pos0, pos1, ..., posN)),
        where pos0, pos1, ..., posN are the positions corresponding to the N axes,
        and all positions & values are shifted with np.fft.ifftshift to put the 0-position in the center.
    '''
    array = np.asanyarray(array)
    if axes is None:
        axes = list(range(array.ndim))
    Ns = [array.shape[ax] for ax in axes]
    df = [df] * len(axes) if not is_iterable(df) else df
    ds = [ds] * len(axes) if not is_iterable(ds) else ds
    x0 = [x0] * len(axes) if not is_iterable(x0) else x0
    if len(df) != len(axes):
        raise InputConflictError(f"len(df)={len(df)} != len(axes)={len(axes)}")
    if len(ds) != len(axes):
        raise InputConflictError(f"len(ds)={len(ds)} != len(axes)={len(axes)}")
    if len(x0) != len(axes):
        raise InputConflictError(f"len(x0)={len(x0)} != len(axes)={len(axes)}")
    shifted_pos = tuple(ifftfreq_shifted(N, df_, rad=rad, x0=x0_, ds=ds_) for N, df_, ds_, x0_ in zip(Ns, df, ds, x0))
    # take inverse fft & shift appropriately.  # [TODO] get the shift / inverse shift correct.
    array_unshifted = np.fft.ifftshift(array, axes=axes)
    vals = np.fft.ifftn(array_unshifted, axes=axes, **kw_np_ifftn)
    return vals, shifted_pos

# [TODO] ifft2, ifft1

def ifftfreq_shifted(N, df=None, *, rad=False, x0=0, ds=None, float_rounding=True):
    '''returns the shifted positions corresponding to the inverse fft of an array of length N.

    df: None or number
        spacing between elements of array (in frequency-space).
        None --> use ds to determine position coordinates instead.
    rad: bool
        whether to interpret frequency-spacing (df) like it is "in radians",
        dividing it by 2 * pi before converting to position-space.
    x0: None or number
        if provided, add value to result such that result[0] == x0.
        (This corresponds to altering position-space coordinates by a constant offset.)
    ds: None or number
        if provided, actually use this as position diff, instead of determining it from d & rad.
    float_rounding: bool
        if True, round ds to a more-likely-to-be-input float.
        e.g. 0.20000000001 --> float(0.2); 0.399999999 --> float(0.4); 0.123456781234 --> unchanged.
    '''
    if df is None and ds is None:
        raise InputMissingError("either df or ds must be provided")
    if df is not None:
        if rad:
            df = df / (2 * np.pi)   # avoid "/=" because it could directly alter the input d.
        ds = 1 / (N * df)
        # could do np.fft.fftfreq, but instead just use known ds & np.arange.
        #pos = np.fft.fftfreq(N, df)   # note, fftfreq is its own inverse.
        #result = np.fft.fftshift(pos)
    if float_rounding:
        ds = math_float_rounding(ds, prec=8)
    result = np.arange(N)
    if x0 is None:  # behave like fftfreq result
        result -= N//2
    result = result * ds
    if x0 is not None:  # choose value for result[0]
        result += x0
    return result



"""
File Purpose: helpful tools for processing data with physical instability in it

E.g. "get t when linear growth ends", "measure growth rate"
"""

import numpy as np

from .instability_xarray_accessor import itAccessor
from ...tools import (
    format_docstring,
    alias,
    XarrayCurveFitter,
)

''' --------------------- fitting --------------------- '''
# (instead of managing theory & theoretical solutions)
# e.g. measure growth rate, t_turb when turbulence begins, saturation level.


def pwl2_flatend(xx, b0, m0, end0):
    '''evaluate xx at piecewise linear function with 2 pieces, with final piece slope=0.
    
    xx: 1D array. Assumed to be monotonically increasing.
    b0: y-intercept of piece 0
    m0: slope of piece 0
    end0: "index" of end of piece 0
        if end0 is not an int, does weighted averaging of xx[int(end0)] and xx[int(end0)+1].
        E.g. end0 = 10.25 --> extend piece 0 to xx[10] + 0.25 * (xx[11] - xx[10])

    b1 is computed based on the other inputs.

    This is a decent approx. for ln(val) with linear growth then saturation.
        xx <--> time
        b0 <--> pre-growth noise level
        m0 <--> growth rate
        end0 <--> "saturation index" when linear growth stops.
        x0 <--> "saturation time" when linear growth stops, where:
                x0 = xx[i0] + (end0 - i0) * (xx[i0+1] - xx[i0], where i0 = int(end0).
        y0  <--> "saturation level"; value when saturated, where:
                y0 = m0 * x0 + b0.
    '''
    # assumptions
    m1 = 0
    # piecewise linear math
    yy0 = m0 * xx + b0
    iend0 = int(end0)
    rounded = end0 - iend0
    x0 = xx[iend0] + rounded * (xx[iend0+1] - xx[iend0])
    y0 = m0 * x0 + b0  # evaluate piece 0 at endpoint.
    b1 = y0 - m1 * x0
    yy1 = m1 * xx + b1
    # add yy * (1 if xindex in section else 0)
    sec0 = np.zeros_like(xx)
    sec0[:iend0+1] = 1  # piece 0 includes the iend0 point
    sec1 = np.zeros_like(xx)
    sec1[iend0+1:] = 1
    return yy0 * sec0 + yy1 * sec1

def pwl3_flatend(xx, b0, m0, end0, m1, end1add):
    '''evaluate xx at piecewise linear function with 3 pieces, with final piece slope=0.

    xx: 1D array. Assumed to be monotonically increasing.
    b0: y-intercept of piece 0
    m0: slope of piece 0
    end0: "index" of end of piece 0
        if end0 is not an int, does weighted averaging of xx[floor(end0)] and xx[ceil(end0)].
        E.g. end0 = 10.25 --> extend piece 0 to xx[10] + 0.25 * (xx[11] - xx[10])
    m1: slope of piece 1
    end1add: "index" of end of piece 1, minus end0.
        end1 = end0 + end1add.  (max=len(xx)-1)
        if end1 is not an int, handle similarly to end0 (see above).

    b2 is computed based on the other inputs.

    This is a decent approx. for ln(val) with linear growth, then damped linear growth, then saturation.
        xx <--> time
        b0 <--> pre-growth noise level
        m0 <--> growth rate of linear growth
        end0 <--> "damped index" when linear growth stops.
        x0 <--> "damped time" when linear growth stops, where:
                x0 = xx[i0] + (end0 - i0) * (xx[i0+1] - xx[i0], where i0 = int(end0).
        y0 <--> pre-damped-growth level, where:
                y0 = m0 * x0 + b0.
                (--> damped growth piece has y-intercept b1 = y0 - m1 * x0.)
        m1 <--> growth rate of damped linear growth
        end1 <--> "saturation index" when damped growth stops, where:
                end1 = end0 + end1add
        x1 <--> "saturation time" when damped growth stops, where:
                x1 = xx[i1] + (end1 - i1) * (xx[i1+1] - xx[i1], where i1 = int(end1).
        y1 <--> "saturation level"; value when saturated, where:
                y21= m1 * x1 + b1.
    '''
    # assumptions
    m2 = 0
    end1 = end0 + end1add
    if end1 > len(xx)-2:
        end1 = len(xx)-2
    # piecewise linear math
    yy0 = m0 * xx + b0
    iend0 = int(end0)
    rounded0 = end0 - iend0
    x0 = xx[iend0] + rounded0 * (xx[iend0+1] - xx[iend0])
    y0 = m0 * x0 + b0  # evaluate piece 0 at endpoint.
    b1 = y0 - m1 * x0
    yy1 = m1 * xx + b1
    iend1 = int(end1)
    rounded1 = end1 - iend1
    x1 = xx[iend1] + rounded1 * (xx[iend1+1] - xx[iend1])
    y1 = m1 * x1 + b1  # evaluate piece 1 at endpoint.
    b2 = y1 - m2 * x1
    yy2 = m2 * xx + b2
    # add yy * (1 if xindex in section else 0)
    sec0 = np.zeros_like(xx)
    sec0[:iend0+1] = 1  # piece 0 includes the iend0 point
    sec1 = np.zeros_like(xx)
    sec1[iend0+1:iend1+1] = 1  # piece 1 includes the iend1 point
    sec2 = np.zeros_like(xx)
    sec2[iend1+1:] = 1
    return yy0 * sec0 + yy1 * sec1 + yy2 * sec2


@itAccessor.register('pwl2_flatend_fitter', totype='array')
@format_docstring(pwl2_flatend_docs=pwl2_flatend.__doc__)
class Pwl2FlatendFitter(XarrayCurveFitter):
    '''CurveFitter with f = pwl2_flatend, for fitting data to piecewise linear with 2 pieces.
    When fitting, use bounds:
        m0 > 0.
        1 <= end0 <= len(xdata)-2

    pwl2_flatend docs copied below, for convenience:
    ------------------------------------------------
    {pwl2_flatend_docs}
    '''
    f = staticmethod(pwl2_flatend)
    pnames = ['b0', 'm0', 'end0']
    pbounds = [
        None,
        (0, None),
        (1, lambda arr, xdim: arr.sizes[xdim]-2)
    ]

    get_xsat = alias('get_x0', doc='''x value at start of saturation. alias to self.get_x0.''')

    def get_x0(self):
        '''return x0, the x value at the end of piece 0.
        (Crashes if run before self.fit())

        x0 = xx[i0] + (end0 - i0) * (xx[i0+1] - xx[i0], where i0 = int(end0).

        [TODO] option to get result +-1 stddev error bounds.
        '''
        xdata = self.xdata
        end0 = self.params.to_dataset('param')['end0']
        iend0 = end0.astype(int)
        round0 = end0 - iend0
        dim = self.dim
        x_at_iend0 = xdata.isel({dim: iend0})
        x_at_iend0_p1 = xdata.isel({dim: iend0+1})
        x0 = x_at_iend0 + round0 * (x_at_iend0_p1 - x_at_iend0)
        return x0

    get_ysat = alias('get_y0', doc='''saturation level (y-value). alias to self.get_y0.''')

    def get_y0(self):
        '''return y0, the y value at the end of piece 0.
        (Crashes if run before self.fit())

        y0 = m0 * x0 + b0.

        [TODO] option to get result +-1 stddev error bounds.
        '''
        x0 = self.get_x0()
        pds = self.params.to_dataset('param')
        m0 = pds['m0']
        b0 = pds['b0']
        y0 = m0 * x0 + b0
        return y0


@itAccessor.register('pwl3_flatend_fitter', totype='array')
@format_docstring(pwl3_flatend_docs=pwl3_flatend.__doc__)
class Pwl3FlatendFitter(XarrayCurveFitter):
    '''CurveFitter with f = pwl3_flatend, for fitting data to piecewise linear with 3 pieces.
    When fitting, use bounds:
        m0, m1 > 0
        1 <= end0 <= len(xdata)-3
        2 <= end1add <= len(xdata)-2

    # [TODO] It would be good to force end0 + end1add < len(xdata)-1. But I don't know how to.

    pwl3_flatend docs copied below, for convenience:
    ------------------------------------------------
    {pwl3_flatend_docs}
    '''
    f = staticmethod(pwl3_flatend)
    pnames = ['b0', 'm0', 'end0', 'm1', 'end1add']
    pbounds = [
        None,
        (0, None),
        (1, lambda arr, xdim: arr.sizes[xdim]-3),
        (0, None),
        (2, lambda arr, xdim: arr.sizes[xdim]-2),
    ]


    def get_x0(self):
        '''return x0, the x value at the end of piece 0.
        (Crashes if run before self.fit())

        x0 = xx[i0] + (end0 - i0) * (xx[i0+1] - xx[i0], where i0 = int(end0).

        [TODO] option to get result +-1 stddev error bounds.
        '''
        xdata = self.xdata
        end0 = self.params.to_dataset('param')['end0']
        iend0 = end0.astype(int)
        round0 = end0 - iend0
        dim = self.dim
        x_at_iend0 = xdata.isel({dim: iend0})
        x_at_iend0_p1 = xdata.isel({dim: iend0+1})
        x0 = x_at_iend0 + round0 * (x_at_iend0_p1 - x_at_iend0)
        return x0

    def get_y0(self):
        '''return y0, the y value at the end of piece 0.
        (Crashes if run before self.fit())

        y0 = m0 * x0 + b0.

        [TODO] option to get result +-1 stddev error bounds.
        '''
        x0 = self.get_x0()
        pds = self.params.to_dataset('param')
        m0 = pds['m0']
        b0 = pds['b0']
        y0 = m0 * x0 + b0
        return y0

    get_xsat = alias('get_x1', doc='''x value at start of saturation. alias to self.get_x1.''')

    def get_x1(self):
        '''return x1, the x value at the end of piece 1.
        (Crashes if run before self.fit())

        x1 = xx[i1] + (end1 - i1) * (xx[i1+1] - xx[i1], where i1 = int(end1),
        and end1 = end0 + end1add.

        [TODO] option to get result +-1 stddev error bounds.
        '''
        xdata = self.xdata
        pds = self.params.to_dataset('param')
        end0 = pds['end0']
        end1add = pds['end1add']
        end1 = end0 + end1add
        iend1 = end1.astype(int)
        round1 = end1 - iend1
        dim = self.dim
        x_at_iend1 = xdata.isel({dim: iend1})
        x_at_iend1_p1 = xdata.isel({dim: iend1+1})
        x1 = x_at_iend1 + round1 * (x_at_iend1_p1 - x_at_iend1)
        return x1

    get_ysat = alias('get_y1', doc='''saturation level (y-value). alias to self.get_y1.''')

    def get_y1(self):
        '''return y1, the y value at the end of piece 1.
        (Crashes if run before self.fit())

        y1 = m1 * x1 + b1.

        [TODO] option to get result +-1 stddev error bounds.
        '''
        x1 = self.get_x1()
        pds = self.params.to_dataset('param')
        m1 = pds['m1']
        b1 = pds['b1']
        y1 = m1 * x1 + b1
        return y1



"""
File Purpose: EppicInputDeck - values from eppic.i file.
"""
import os

import numpy as np

from .eppic_dimensions import (
    EppicDist, EppicDistList,
    EppicNeutral, EppicNeutralList,
)
from .eppic_io_tools import read_eppic_i_file, infer_eppic_dist_names
from ...units import UnitsManagerPIC
from ...tools import (
    alias,
    DictlikeFromKeysAndGetitem,
)
from ...defaults import DEFAULTS


''' --------------------- EppicInputDeck --------------------- '''

class EppicInputDeck(DictlikeFromKeysAndGetitem):
    '''info from eppic.i file, in a class.
    Use EppicInputDeck.from_file() to create an instance from a file.
    units are "raw". Numerical values equal those from the eppic.i file.

    params: dict
        values of params
    distributions: dict or EppicDistList
        dict of {distribution number: EppicDist}, or an already-created EppicDistList.
        if dict, will be internally stored as EppicDistList.from_dict(distributions), instead.
    filename: None or str
        path to eppic.i file this came from. E.g. 'path/to/eppic.i'
        str will be internally stored as os.path.abspath(filename), instead.

    Examples:
        eid = EppicInputDeck.from_file('eppic.i')
        eid[1]          # EppicDist for distribution 1
        eid['nz']       # number of grid points in z direction
        eid.params      # dict of all global params
        eid.dists       # EppicDistList of all the distributions
        eid.dists.get('e')    # get EppicDist with name 'e'
        eid.dists.get(0)      # get EppicDist number 0 from this list
        eid[2]['m']           # mass parameter for distribution 2
        eid[0]['coll_rate']   # collision rate parameter for distribution 0
        eid.keys()      # available keys
        eid.items()     # available (key, value) pairs
        eid[0].keys()   # available keys for distribution 0
        eid.get('k', 7) # eid['k'] if it exists, else 7.
    '''
    dist_type = EppicDist  # type for distributions
    dlist_type = EppicDistList  # type for list of distributions
    neutral_type = EppicNeutral  # type for neutral fluid
    nlist_type = EppicNeutralList  # type for list of neutral fluids

    # # # CREATION / INITIALIZATION # # #
    def __init__(self, params, distributions, *, filename=None, **kw_super):
        self.params = params
        self.distributions = distributions if isinstance(distributions, self.dlist_type) \
                                else self.dlist_type.from_dict(distributions)
        self._init_neutral_fluid()
        self.filename = None if filename is None else os.path.abspath(filename)
        super().__init__(**kw_super)

    dists = alias('distributions')

    @classmethod
    def from_file(cls, filename='eppic.i', *, dist_names=None, **kw__init):
        '''return EppicInputDeck from eppic.i file.
        filename: str
            path to eppic.i file.
            abspath(filename) will be saved to result, for easier bookkeeping.
        dist_names: None or dict
            {N: name for distribution N} for any number of distributions in filename.
            None --> infer from file.
        '''
        params, dists = read_eppic_i_file(filename)
        if dist_names is None:
            dist_names = infer_eppic_dist_names(filename)
        if dist_names is not None:
            for N, name in dist_names.items():
                if name is not None:
                    dists[N][f'name{N}'] = name    # (the {N} in the key is required, then removed, by EppicDist.from_dict)
        dist_dict = cls.dist_type.dists_from_dict(dists)
        return cls(params, dist_dict, filename=filename, **kw__init)

    def _init_neutral_fluid(self):
        '''sets self.neutral and self.neutrals based on params of self.
        Particularly, uses params['m_neutral'].
        '''
        m = self.params['m_neutral']
        vth = self.params.get('vth_neutral', None)
        self.neutral = self.neutral_type(m, vth=vth)
        self.neutrals = self.nlist_type([self.neutral])

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [f'with {len(self.dists)} distributions']
        if self.filename is not None:
            contents.append(f'filename={self.filename!r}')
        contents_str = ', '.join(contents)
        return f'{type(self).__name__}({contents_str})'

    # # # GETTING VALUES & ITERATING # # #
    def __getitem__(self, key):
        '''return self.params[key] if key in self.params, else self.dists[key].'''
        try:
            return self.params[key]
        except KeyError:
            pass  # handled below, to make clearer error message if failing.
        try:
            return self.dists[key]
        except (KeyError, IndexError, ValueError):
            errmsg = f'key {key!r} not in {type(self).__name__}.params or .dists'
            raise KeyError(errmsg) from None

    def keys(self):
        '''return tuple of keys to access self.dists or self.params'''
        return tuple(key for key in (*range(len(self.dists)), *self.params.keys()))

    # # # FILE PATH # # #
    @property
    def dirname(self):
        '''os.path.dirname(self.filename).
        if self.filename is None, raise FileNotFoundError.
        '''
        filename = self.filename
        if filename is None:
            raise FileNotFoundError('self.filename is None; cannot determine dirname.')
        return os.path.dirname(filename)

    # # # DIMENSIONS # # #
    def maindims(self):
        '''returns tuple of main dimensions in this input deck.
        inferred from self['nx'], 'ny', 'nz', and 'ndim_space'
        '''
        ndim_space = self['ndim_space']
        if ndim_space == 3:
            return ('x', 'y', 'z')
        elif ndim_space == 2:
            nx = self.get('nx', None)
            if nx is None: return ('y', 'z')
            ny = self.get('ny', None)
            if ny is None: return ('x', 'z')
            nz = self.get('nz', None)
            if nz is None: return ('x', 'y')
            # provided nx, ny, and nz...
            if nz == 1: return ('x', 'y')
            elif ny == 1: return ('x', 'z')
            elif nx == 1: return ('y', 'z')
            # nx, ny, and nz were all provided, and not 1.
            if getattr(self, 'verbose', True) and not getattr(self, 'warned_ndim2', False):
                print("warning, ndim_space=2 but input deck contains 'nx', 'ny', and 'nz', all != 1.")
                print("assuming space dims ('x', 'y').")
                self.warned_ndim2 = True  # only print the warning once per input_deck.
            return ('x', 'y')  # by default, return ('x', 'y').
        else:
            raise NotImplementedError(f'{type(self).__name__}.maindims() when ndim_space={ndim_space}')

    def _get_nspace_simple(self, xstr):
        '''get "simple" number of cells in {x} dimension. == n{x} // nout_avg
        xstr should be 'x', 'y', or 'z'.
        Note that for x='x', actual number of cells may be affected by nsubdomains.
        See also: self._get_nspace.
        '''
        return self.get(f'n{xstr}', 1) // self.get('nout_avg', 1)

    def _get_velocity_nspace_simple(self, xstr, N):
        '''get "simple" number of cells in {vx} dimension. == pn{vx}{N}
        xstr should be 'vx', 'vy', or 'vz'.
        N is the distribution number.
        Note that for x='x', actual number of cells may be affected by nsubdomains.
        See also: self.get_nspace.
        '''
        N = str(N)
        return self.get(f'pn{xstr}{N}', 1)

    def get_nspace(self, xstr):
        '''get number of cells in {x} dimension, accounting for nout_avg and nsubdomains.
        See also: Nx, Ny, Nz.
        '''
        result = self._get_nspace_simple(xstr)
        if xstr == 'x':
            result = result * self.get('nsubdomains', 1)
        return result

    Nx = property(lambda self: self.get_nspace('x'),
                  doc='''actual total number of grid cells in the x dimension.''')
    Ny = property(lambda self: self.get_nspace('y'),
                  doc='''actual total number of grid cells in the y dimension.''')
    Nz = property(lambda self: self.get_nspace('z'),
                  doc='''actual total number of grid cells in the z dimension.''')

    def get_velocity_nspace(self, xstr, N):
        '''get number of cells in {vx} dimension.
        N is the distribution number.
        See also: Nvx, Nvy, Nvz.
        '''
        result = self._get_velocity_nspace_simple(xstr, N)
        return result

    Nvx = property(lambda self: [self.get_velocity_nspace('x', N) for N in range(len(self.dists))],
                  doc='''list of actual total numbers of grid cells in the vx dimension (ordered by distribution).''')
    Nvy = property(lambda self: [self.get_velocity_nspace('x', N) for N in range(len(self.dists))],
                  doc='''list of actual total numbers of grid cells in the vy dimension (ordered by distribution).''')
    Nvz = property(lambda self: [self.get_velocity_nspace('x', N) for N in range(len(self.dists))],
                  doc='''list of actual total numbers of grid cells in the vz dimension (ordered by distribution).''')

    def get_dspace(self, xstr):
        '''get d{x} [raw units] between output cells.
        accounts for nout_avg; see also: Dx, Dy, Dz.
        if {x} not provided in self (e.g. 2D in x & y, xstr='z'), returns 1 * nout_avg.
        '''
        return self.get(f'd{xstr}', 1) * self.get('nout_avg', 1)

    Dx = property(lambda self: self.get_dspace('x'),
                    doc='''dx between adjacent output cells [raw units]''')
    Dy = property(lambda self: self.get_dspace('y'),
                    doc='''dy between adjacent output cells [raw units]''')
    Dz = property(lambda self: self.get_dspace('z'),
                    doc='''dz between adjacent output cells [raw units]''')

    def _get_velocity_extents(self, xstr, N):
        '''get (p{vx}min{N}, p{vx}max{N}) [raw units].
        N is the distribution number.
        See also: Dvx, Dvy, Dvz.
        '''
        return (self.get(f'p{xstr}min{N}', 1), self.get(f'p{xstr}max{N}', 1))

    def get_coord(self, xstr):
        '''get coordinate array [raw units] for this dimension.
        xstr: str
            dimension name. 'x', 'y', or 'z'.
        '''
        dx = self.get_dspace(xstr)
        Nx = self.get_nspace(xstr)
        return np.arange(Nx) * dx

    def get_space_coords(self):
        '''return dict of {x: xcoord array} for x in self.maindims(), a subset of ('x', 'y', 'z').'''
        return {x: self.get_coord(x) for x in self.maindims()}

    def get_dspace_min(self):
        '''return min of self.Dx, self.Dy, self.Dz.'''
        return min(self.Dx, self.Dy, self.Dz)

    def get_dspace_max(self):
        '''return max of self.Dx, self.Dy, self.Dz.'''
        return max(self.Dx, self.Dy, self.Dz)

    # # # UNITS # # #
    def get_units_manager(self, *, u_l=None, u_t=None, u_n=None, ne_si=None, **kw_u_from_pic):
        '''return UnitsManagerPIC for this input deck. Assumes distribution 0 is electrons.
        Must provide u_l, u_t, u_n, or ne_si.
        Will read ne, qe, me, and eps0 from the input deck, unless they are provided in kw_u_from_pic.
        See help(UnitsManagerPIC.pic_ambiguous_unit) for details.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        if 'ne' not in kw_u_from_pic: kw_u_from_pic['ne'] = self[0]['n0']
        if 'qe' not in kw_u_from_pic: kw_u_from_pic['qe'] = abs(self[0]['q'])
        if 'me' not in kw_u_from_pic: kw_u_from_pic['me'] = self[0]['m']
        if 'eps0' not in kw_u_from_pic: kw_u_from_pic['eps0'] = self['eps']
        return UnitsManagerPIC.from_pic(u_l=u_l, u_t=u_t, u_n=u_n, ne_si=ne_si, **kw_u_from_pic)

    # # # QUASINEUTRAL? # # #
    def is_quasineutral(self):
        '''return whether this input deck is in quasineutral mode.
        as a proxy, guessing quasineutral iff any dist.is_hybrid() for dist in self.dists.
        '''
        return any(dist.is_hybrid() for dist in self.dists)




"""
File Purpose: ionization-related quantities (e.g. ionization fraction)
This includes an implementation of the saha ionization equation.
"""

import numpy as np
import xarray as xr

from .mhd_genrad_tables import GenradTableManager
from .mhd_fluids import Specie
from ..elements import Element
from ...defaults import DEFAULTS
from ...dimensions import SINGLE_FLUID
from ...errors import (
    FluidValueError, FluidKeyError,
    LoadingNotImplementedError,
)
from ...quantities import QuantityLoader
from ...tools import (
    simple_property,
    format_docstring,
    xarray_promote_dim,
)


''' --------------------- Ionization Helper Methods --------------------- '''

_paramdocs_saha = {
    'ne': '''number or array
        number density of electrons.''',
    'T': '''number or array
        temperature.''',
    'xi': '''number
        first ionization potential.''',
    'g1g0': '''number
        ratio of element's g (degeneracy of states) for g1 (ions) to g0 (neutrals).''',
    'u': '''None or UnitsManager
        units to use; determines expected units system for input & output.
        Also will grab relevant physical constants (such as kB) directly from u.
        None --> make new UnitsManager with SI units.''',
    'saha_equation': ''''(n1/n0) = (1/ne) * (2.0 / lde^3) * (g1 / g0) * exp(-xi / (kB * T))
        where the terms are defined as follows:
            T : temperature
            ne: electron number density
            n1: number density of element's once-ionized ions.
            n0: number density of element's neutrals.
            g1: "degeneracy of states" for element's once-ionized ions.
            g0: "degeneracy of states" for element's neutrals.
            xi: element's first ionization energy.
            lde: electron thermal deBroglie wavelength:
                lde^2 = hplanck^2 / (2 pi me kB T)''',
}

# note: saha_n1n0 not used internally... provided for reference / possible convenience.
@format_docstring(**_paramdocs_saha)
def saha_n1n0(*, ne, T, xi, g1g0, u=None):
    '''return (n1/n0) for an element, via saha equation.

    ne: {ne}
    T: {T}
    xi: {xi}
    g1g0: {g1g0}
    u: {u}

    SAHA_EQUATION:
        {saha_equation}
    '''
    if u is None:
        u = UnitsManager()
    # [EFF] calculate (2.0 / lde^3) ignoring T, to avoid extra array multiplications.
    ldebroge_constant = 2.0 / (u('hplanck')**2 / (2 * np.pi * u('me') * u('kB')))**(3/2)
    ldebroge_factor = ldebroge_constant * T**(3/2)
    result = (ldebroge_factor / ne) * (g1g0 * np.exp(-xi / (u('kB') * T)))
    return result


''' --------------------- MhdMultifluidIonizationLoader --------------------- '''

class MhdMultifluidIonizationLoader(QuantityLoader):
    '''ionization-related quantities (e.g. ionization fraction)
    Includes an implementation of the saha ionization equation.
    [TODO] include more ionization-related quantities, such as rates?
    '''
    # # # IONFRAC GENERIC / DISPATCH # # #
    cls_behavior_attrs.register('ionfrac_mode', default='best_no_radtab')
    ionfrac_mode = simple_property('_ionfrac_mode',
        default='best_no_radtab',  # because I'm not convinced radtab is any better than saha (-SE, 2025/01/31)
        validate_from='IONFRAC_MODE_OPTIONS',
        doc='''mode for calculating self('ionfrac').
        ignored when self.fluid is SINGLE_FLUID. See self.IONFRAC_MODE_OPTIONS for options.''')
    IONFRAC_MODE_OPTIONS = {
        'best': '''use 'n' if self.fluid is an ion Specie, else 'best_no_n'.''',
        'best_no_radtab': '''use 'n' if self.fluid is an ion Specie, else 'saha'.''',
        'n': '''ionfrac = n / n_elem, and fluid must be an ion Specie ([TODO] or IonMixture?).
            [EFF] use 'saha' if self.fluid ntype is 'elem' or 'saha'
                (computes saha_n1n0, and ionfrac_saha from that).
                Otherwise, directly compute nII and n_elem, and take their ratio
                (ensures consistency e.g. if nII is coming from aux not saha).''',
        'best_no_n': '''best mode without getting n directly. 'radtab' if available else 'saha'.''',
        'saha': '''use saha ionization equation to calculate ionfrac.''',
        'radtab': '''lookup from self.radtab.''',
    }

    @known_var(load_across_dims=['fluid'])
    def get_ionfrac_type(self):
        '''ionfrac_type for self.fluid; affects 'ionfrac' result.
        The output array values will each be one of:
            'SINGLE_FLUID' <--> ionfrac = ne / n.
            'n' <--> ionfrac from n / n_elem, and fluid must be an ion Specie.
            'saha' <--> ionfrac from saha equation
            'radtab' <--> ionfrac based on self.radtab table lookup
            'nan' <--> ionfrac not defined for this fluid
        '''
        f = self.fluid
        if f is SINGLE_FLUID:
            result = 'SINGLE_FLUID'
        elif f.is_electron():
            result = 'nan'
            self._handle_typevar_nan(errmsg=f"ionfrac_type, when fluid={self.fluid!r}")
        else:
            result = self.ionfrac_mode
            if result == 'best':
                if isinstance(f, Specie) and f.is_ion():
                    result = 'n'
                else:
                    result = 'best_no_n'
            if result == 'best_no_radtab':
                if isinstance(f, Specie) and f.is_ion():
                    result = 'n'
                else:
                    result = 'saha'
            if result == 'best_no_n':
                if self.radtab.neufrac_available(f):
                    result = 'radtab'
                else:
                    result = 'saha'
            if result == 'n':
                # use saha if appropriate.
                ntype = self('ntype').item()
                if ntype == 'saha':
                    result = 'saha'
        return xr.DataArray(result)

    _IONFRAC_TYPE_TO_IONFRAC_DEPS = {
        'SINGLE_FLUID': ['ne', 'SF_n'],
        'n': ['n', 'n_elem'],
        'saha': ['ionfrac_saha'],
        'radtab': ['ionfrac_radtab'],
    }
    @known_var(partition_across_dim=('fluid', 'ionfrac_type'), partition_deps='_IONFRAC_TYPE_TO_IONFRAC_DEPS')
    def get_ionfrac(self, *, ionfrac_type):
        '''ionization fraction(s) of element(s) of self.fluid
        if SINGLE_FLUID, return ne / n.
            where ne = electron number density, n = total number density for all elements, excluding electrons.
            assumes quasineutrality, and that only once-ionized ions are relevant --> sum_ions(nion) = ne.
        else if self.ionfrac_mode implies 'n', return n / n_elem.
        else if self.ionfrac_mode implies 'saha', return 1 / (1 + 1/saha_n1n0).
        else if self.ionfrac_mode implies 'radtab', return ionfrac based on self.radtab table lookup from self('T')
        '''
        if ionfrac_type == 'SINGLE_FLUID':
            if self.current_n_fluid() > 1:
                raise LoadingNotImplementedError('[TODO] get_ionfrac with multiple SINGLE_FLUID...')
            with self.using(fluid=SINGLE_FLUID):
                result = self.assign_fluid_coord(self('ne/n'), overwrite=True)
        elif ionfrac_type == 'n':
            result = self('n') / self('n_elem')
        elif ionfrac_type == 'saha':
            result = self('ionfrac_saha')
        elif ionfrac_type == 'radtab':
            result = self('ionfrac_radtab')
        elif ionfrac_type == 'nan':
            result = self('nan')
        else:
            raise LoadingNotImplementedError(f'ionfrac_type={ionfrac_type!r} not yet supported.')
        return result

    @known_var(deps=['ionfrac'])
    def get_neufrac(self):
        '''neutral fraction(s) of element(s) of self.fluid. neufrac = 1 - ionfrac.'''
        return 1 - self('ionfrac')

    @known_var(deps=['ionfrac', 'n_elem'])
    def get_nII(self):
        '''number density of once-ionized species of element(s) of self.fluid. nII = n * ionfrac
        see help(self.get_ionfrac) for more details.
        '''
        return self('n_elem') * self('ionfrac')

    @known_var(deps=['neufrac', 'n_elem'])
    def get_nI(self):
        '''number density of neutral species of element(s) of self.fluid. nI = n * neufrac == n * (1 - ionfrac)
        see help(self.get_saha_n1n0) and help(self.get_ionfrac) for more details.
        assumes only once-ionized ions are relevant (ignore twice+ ionized ions).
        '''
        return self('n_elem') * self('neufrac')

    @known_var(load_across_dims=['fluid'])
    def get_n_from_ionfrac_charge__type(self):
        '''n_from_ionfrac_charge__type for self.fluid, for inferring densities from ionization fraction.
            'I' <--> neutrals
            'II' <--> once-ionized ions
            'III+' <--> twice+ ionized ions
            'nan' <--> unknown charge, or electrons (getting ne from ionfrac not possible here).
        '''
        q = getattr(self.fluid, 'q', None)
        if (q is None) or q < 0:
            result = 'nan'
        elif q == 0:
            result = 'I'
        elif q == 1:
            result = 'II'
        else:
            assert q > 1
            result = 'III+'
        if result == 'nan':
            self._handle_typevar_nan(errmsg=f"n_from_ionfrac_charge__type, when fluid={self.fluid!r}")
        return xr.DataArray(result)

    # # # SAHA IONIZATION EQUATION # # #
    @known_var(deps=['T'])
    def get_ldebroge(self):
        '''electron thermal deBroglie wavelength.
        lde^2 = hplanck^2 / (2 pi me kB T)
        '''
        T = self('T')
        const = np.sqrt(self.u('hplanck')**2 / (2 * np.pi * self.u('me') * self.u('kB')))
        return const / np.sqrt(T)

    @known_var(deps=['SF_T'], ignores_dims=['fluid'])
    def get_saha_factor_ldebroge(self, *, _T=None):
        '''(2.0 / lde^3) for saha equation. (See help(self.get_saha_n1n0) for full saha equation.)
        [EFF] computed "efficiently", i.e. combine all constants before including T contribution.
        [EFF] for efficiency, can provide (single fluid) T if already known.
        '''
        ldebroge_constant = 2.0 / (self.u('hplanck')**2 / (2 * np.pi * self.u('me') * self.u('kB')))**(3/2)
        T = self('SF_T') if _T is None else _T
        return ldebroge_constant * T**(3/2)

    @known_var(deps=['ne', 'saha_factor_ldebroge'])
    def get_saha_factor_ldebroge_ne(self, *, _T=None):
        '''(ldebroge_factor / ne) for saha equation. (See help(self.get_saha_n1n0) for full saha equation.)
        [EFF] for efficiency, can provide (single fluid) T if already known.
        '''
        return self('saha_factor_ldebroge', _T=_T) / self('ne')

    @known_var(load_across_dims=['fluid'])
    def get_saha_g1g0(self):
        '''ratio of self.fluid element's g (degeneracy of states) for g1 (ions) to g0 (neutrals).'''
        f = self.fluid
        if f is SINGLE_FLUID:
            raise FluidValueError('get_saha_g1g0 requires self.fluid correspond to element(s), not SINGLE_FLUID.')
        elif isinstance(f, (Element, Specie)):
            return xr.DataArray(f.get_element().saha_g1g0)
        else:
            raise NotImplementedError(f'fluid of type {type(f)} not yet supported for get_saha_g1g0.')

    @known_var(load_across_dims=['fluid'], aliases=['first_ionization_energy'])
    def get_ionize_energy(self):
        '''self.fluid element's first ionization energy.'''
        f = self.fluid
        if f is SINGLE_FLUID:
            raise FluidValueError('get_ionize_energy requires self.fluid correspond to element(s), not SINGLE_FLUID.')
        elif isinstance(f, (Element, Specie)):
            return xr.DataArray(f.get_element().ionize_ev * self.u('eV'))
        else:
            raise NotImplementedError(f'fluid of type {type(f)} not yet supported for get_ionize_energy.')

    @known_var(deps=['ionize_energy', 'SF_T'])
    def get_saha_factor_exp(self, *, _T=None):
        '''exp(-xi / (kB * T)) for saha equation. (See help(self.get_saha_n1n0) for full saha equation.)
        xi = first ionization energy.
        [EFF] for efficiency, can provide (single fluid) T if already known.
        '''
        T = self('SF_T') if _T is None else _T
        return np.exp(-self('ionize_energy') / (self.u('kB') * T))

    @known_var(deps=['saha_factor_ldebroge_ne', 'saha_g1g0', 'saha_factor_exp'])
    @format_docstring(**_paramdocs_saha, sub_ntab=1)
    def get_saha_n1n0(self):
        '''(n1/n0) for an element, via saha equation:
            {saha_equation}
        '''
        T = self('T', fluid=SINGLE_FLUID)
        f_ldebroge_ne = self('saha_factor_ldebroge_ne', _T=T)
        f_g1g0 = self('saha_g1g0')
        f_exp = self('saha_factor_exp', _T=T)
        return f_ldebroge_ne * f_g1g0 * f_exp

    @known_var(deps=['saha_n1n0'])
    def get_ionfrac_saha(self):
        '''ionization fraction based on saha equation: ionfrac_saha = 1 / (1 + 1 / saha_n1n0)

        Equivalent: n1 / (n1 + n0), where n1 & n0 = number density for element's ions (n1) & neutrals (n0).
            assumes only once-ionized ions are relevant (i.e., n0 + n1 = element's total number density).
        '''
        return 1 / (1 + 1 / self('saha_n1n0'))

    @known_var(deps=['ionfrac_saha'])
    def get_neufrac_saha(self):
        '''neutral fraction based on saha ionization. neufrac_saha = 1 - ionfrac_saha.'''
        return 1 - self('ionfrac_saha')

    @known_var(deps=['ionfrac_saha', 'n_elem'], aliases=['saha_n1'])
    def get_nII_saha(self):
        '''number density of once-ionized species of element(s) of self.fluid. nII = n * ionfrac_saha
        assumes only once-ionized ions are relevant (ignore twice+ ionized ions).
        '''
        with self.using(ionfrac_mode='saha'):
            return self('nII')

    @known_var(deps=['ionfrac_saha', 'n_elem'], aliases=['saha_n0'])
    def get_nI_saha(self):
        '''number density of neutral species of element(s) of self.fluid. nI = n * (1 - ionfrac_saha)
        assumes only once-ionized ions are relevant (ignore twice+ ionized ions).
        '''
        with self.using(ionfrac_mode='saha'):
            return self('nI')

    @known_var(partition_across_dim=('fluid', 'n_from_ionfrac_charge__type'),
               partition_deps={'I': 'nI_saha', 'II': 'nII_saha', 'III+': '0', 'nan': 'nan'})
    def get_n_saha(self, *, n_from_ionfrac_charge__type):
        '''number density self.fluid, based on saha ionization.
        result depends on n_from_ionfrac_charge__type, which is based on fluid.q.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        type_to_var = {'I': 'nI_saha', 'II': 'nII_saha', 'III+': '0', 'nan': 'nan'}
        qtype = n_from_ionfrac_charge__type
        if qtype in type_to_var:
            var = type_to_var[qtype]
            return self(var)
        else:
            raise LoadingNotImplementedError(f'n_from_ionfrac_charge__type={qtype!r} not yet supported.')

    # # # GENRAD # # #
    radtab_cls = GenradTableManager  # class for making radtab during self._default_radtab()

    radtab = simple_property('_radtab', setdefaultvia='_default_radtab',
        doc='''dict-like manager for Genrad tables.
        Each table=radtab[var] should have a table.interp(values) method,
            which returns value of var given appropriate values.
            See table.invar for name of expected inputs.''')

    def _default_radtab(self):
        '''return default value of self.radtab: GenradTableManager.from_defaults().'''
        return self.radtab_cls.from_defaults()

    @known_var(deps=['SF_T'])
    def get_neufrac_radtab(self):
        '''neutral fraction based on self.radtab and self('SF_T').
        crash if radtab not applicable to all fluids in self.fluid.
        '''
        T = self('SF_T')
        elems = [f.get_element() for f in self.fluid_list()]
        for f in self.fluid_list():
            if not self.radtab.neufrac_available(f):
                available = list(self.radtab.neufrac_available().keys())
                errmsg = (f'neufrac_radtab not available for all elements in self.fluid. '
                          f'Available: {available!r}; requested: {[str(e) for e in elems]}')
                if self.call_depth > 1 and self.ionfrac_mode == 'radtab':
                    errmsg += "\nConsider using a different ionfrac_mode, e.g. 'best' or 'saha'."
                raise FluidKeyError(errmsg)
        squeeze_later = (not self.fluid_is_iterable())
        result = []
        for f in self.iter_fluid():
            neufrac = self.radtab.neufrac_table(f).interp(T)
            neufrac = self.assign_fluid_coord(neufrac, overwrite=True)
            result.append(neufrac)
        result = self.join_fluids(result)
        if squeeze_later:
            result = xarray_promote_dim(result, 'fluid').squeeze('fluid')
        return result

    @known_var(deps=['neufrac_radtab'])
    def get_ionfrac_radtab(self):
        '''ionization fraction based on self.radtab. ionfrac_radtab = 1 - neufrac_radtab.
        crash if radtab not applicable to all fluids in self.fluid.
        '''
        return 1 - self('neufrac_radtab')

    @known_var(deps=['ionfrac_radtab', 'n_elem'])
    def get_nII_radtab(self):
        '''number density of once-ionized species of element(s) of self.fluid. nII = n * ionfrac_radtab'''
        with self.using(ionfrac_mode='radtab'):
            return self('nII')

    @known_var(deps=['ionfrac_radtab', 'n_elem'])
    def get_nI_radtab(self):
        '''number density of neutral species of element(s) of self.fluid. nI = n * neufrac_radtab'''
        with self.using(ionfrac_mode='radtab'):
            return self('nI')

    @known_var(partition_across_dim=('fluid', 'n_from_ionfrac_charge__type'),
               partition_deps={'I': 'nI_radtab', 'II': 'nII_radtab', 'III+': '0', 'nan': 'nan'})
    def get_n_radtab(self, *, n_from_ionfrac_charge__type):
        '''number density self.fluid, based on self.radtab lookups of self('SF_T')
        result depends on n_from_ionfrac_charge__type, which is based on fluid.q.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        type_to_var = {'I': 'nI_radtab', 'II': 'nII_radtab', 'III+': '0', 'nan': 'nan'}
        qtype = n_from_ionfrac_charge__type
        if qtype in type_to_var:
            var = type_to_var[qtype]
            return self(var)
        else:
            raise LoadingNotImplementedError(f'n_from_ionfrac_charge__type={qtype!r} not yet supported.')



"""
Package purpose: convenient plotting methods & misc plotting tools
"""
# module references, in case other packages need to access internal modules.
# provide module references for all modules which might be overwritten by same-named object.
# E.g. from .subplots import subplots --> "subplots" here will be the function not the module.
from . import subplots as _subplots_module
from .plot_tools import _colorbar_module

from .contours import (
    XarrayContourPlotElement, XarrayContour,
)
from .faceplot import (
    FaceplotPlotElement, Faceplot,
    xarray_faces_3D,
)
from .images import (
    ImagePlotElement, Image,
    XarrayImagePlotElement, XarrayImage,
)
from .labels import (
    TextPlotElement, MovieText, XarrayText,
    xarray_title_plot_node, xarray_suptitle_plot_node,
    title_from_coords,
    XarraySubplotTitlesInferer,
)
from .lines import (
    XarrayLinePlotElement, XarrayLine,
)
from .movies import (
    FuncAnimation,
    MoviePlotElement, MoviePlotNode,
    EmptyMovieNode, MovieOrganizerNode,
)
from .patches import (
    PatchPlotElement, XarrayPatch,
    RectanglePatchPlotElement, XarrayRectanglePatch,
    LimsPatchPlotElement, XarrayLimsPatch,
)
from .plot_settings import (
    PlotSettings, PlotSetting,
    DEFAULT_PLOT_SETTINGS, MPL_KWARGS,
)
from .plotter_manager import MetaPlotterManager, PlotterManager
from .scatter import (
    ScatterPlotElement,
    XarrayScatter,
)
from .subplots import (
    subplots_figsize,
    subplots, Subplots,
)
from .subplots_extensions import (
    MovieSubplots,
    XarraySubplots,
)
from .xarray_timelines import (
    XarrayTimelines,
    IndexableCycler,
)

# PLOT_TOOLS SUBPACKAGE #
from .plot_tools import (
    # colorbar
    make_cax, find_mappable, colorbar,
    Colorbar,
    # colors
    get_cmap,
    ColormapExtremes, BaseColormap, Colormap, cmap,
    CMAPS,
    # currently_active
    current_figure_exists, current_figure_or_None,
    current_axes_exists, current_axes_or_None, current_axes_has_data,
    current_image_exists, current_image_or_None,
    maintaining_current_plt, maintaining_current_figure, maintaining_current_axes, maintaining_current_image,
    using_current_plt, using_current_figure, using_current_axes, using_current_image,
    MaintainingCurrentPlt, UsingCurrentPlt,
    # lims
    calculate_vlims, update_vlims, calculate_lims_from_margin,
    get_data_interval, get_lims_with_margin,
    plt_zoom, plt_zoomx, plt_zoomy, set_margins_if_provided,
    set_lims_if_provided,
    # plot_dims
    PlotDimsMixin,
    infer_movie_dim, infer_xy_dims, infer_xyt_dims,
    # plot_tricks
    ax_outline, fig_outline, ax_remove_ticks, ax_remove_ticklabels,
    get_colorbar_axes,
    get_min_n_ticks, set_min_n_ticks,
    use_simple_log_tick_locator, SimpleLogTickLocator, use_simple_ticks_renamer,
    ax_aspect,
    plt_transformer,
    PLOT_LOCATION_NAMES, plot_locations, plot_note,
)


"""
Package Purpose: collision frequency calculations
"""
from .collisions_cross_sections_loader import CollisionsCrossSectionsLoader
from .collisions_loader import CollisionsLoader
from .cross_section_tools import CrossTable



"""
File Purpose: MultiSlices class

[TODO] encapsulate code shared between here and FFTSlices?
[TODO] record coordinate at ikeep, e.g. keep_x_y array should have coordinate z=...
"""
import itertools

from ...tools import (
    UNSET,
    format_docstring,
    simple_property,
)

_paramdocs_multi_slices = {
    'ndim': '''None or int
        None --> ignore, and do not create special slices.
        int --> create special slices to keep this many dims after applying each slice.
        Example: MultiSlices(ndim=2) is shorthand for
            "MultiSlices with one slices for every possible combination of keeping 2 dims".
        Example: MultiSlices(ndim=2, dims=['x', 'y', 'z'], ikeep=0) is equivalent to:
            MultiSlices(keep_x_y=dict(z=0), keep_y_z=dict(x=0), keep_x_z=dict(y=0))
        Example: MultiSlices(ndim=1, dims=['x', 'y', 'z'], ikeep=0) is equivalent to:
            MultiSlices(keep_x=dict(y=0, z=0), keep_y=dict(x=0, z=0), keep_z=dict(x=0, y=0))''',
    'dims': '''None, iterable of str, or callable of 0 arguments -> iterable of str.
        dimensions to consider when applying special slices.
        None --> do not create any special slices.''',
    'ikeep': '''int or number between -1 < ikeep < 1
        index to take when picking a single value for sliced dimensions for special slices.
        Default is 0, e.g. when slicing x, keep x[0].
        int --> when slicing dim, keep dim[ikeep]. E.g. 10 --> keep x[10]
        non-int between -1 and 1 --> multiply by length of dim to get index.
                                    see interprets_fractional_indexing for more details.''',
}


@format_docstring(**_paramdocs_multi_slices, sub_ntab=2)
class MultiSlices(dict):
    '''managing multiple slices dicts in one place, while also providing options for "special slices".
    Special slices options involve creating slices dicts for each possible combination of N dimensions,
        e.g. in a 3D simulation with dimensions ['x', 'y', 'z'], creating slices for N=2 means
        slices would be dict(keep_x_y=dict(z=0), keep_y_z=dict(x=0), keep_x_z=dict(y=0)).

    keys should be ANY names; these will be associated with results from applying each slice.
        (in MainDimensionsHaver, keys tell names of data_vars in resulting xarray.Dataset)
    values should be a dicts of slices or other indexers.
        e.g. one value might be {{'x': slice(0, 10), 'y': 5, 'z': 0}}.
        (in MainDimensionsHaver, these correspond to any valid value for obj.slices)

    multi_slices: dict
        can provide keys and values here if desired, instead of directly as kwargs.

    The following can be provided during __init__ or set as attrs of self:
        SPECIAL SLICES OPTIONS:
            ndim: {ndim}
            dims: {dims}
            ikeep: {ikeep}

            if ndim and dims are provided, special slices will be included in self.get_slices().
            Note that the naming convention for keys will be 'keep_'+'_'.join(unsliced_dims),
                e.g. 'keep_x', 'keep_x_y', or 'keep_x_y_z'.
                unsliced_dims will be in the same order as dims.
    '''
    # # # CREATION / INITIALIZATION # # #
    def __init__(self, *args_super, multi_slices=dict(), ndim=None, dims=None, ikeep=0, **kw_super):
        super().__init__(*args_super, **multi_slices, **kw_super)
        self.ndim = ndim
        self.dims = dims
        self.ikeep = ikeep

    _kw_init = ('ndim', 'dims', 'ikeep')

    def as_kw(self):
        '''return dict of kwargs to use to recreate this object.'''
        kw = {k: getattr(self, k) for k in self._kw_init}
        kw['multi_slices'] = self
        return kw

    def copy(self):
        '''return a copy of self'''
        return type(self)(**self.as_kw())

    # # # PROPERTIES # # #
    dims = simple_property('_dims',
            doc='''dims to consider when applying special slices.
            if set to be a callable, will call it each time getting self.dims.''')
    @dims.getter
    def dims(self):
        return self._dims() if callable(self._dims) else self._dims

    # # # GETTING SLICES # # #
    @format_docstring(**_paramdocs_multi_slices, sub_ntab=2)
    def get_slices(self, dims=UNSET, *, ikeep=UNSET):
        '''return dict of all multi_slices to apply.
        if self.ndim is None, only use the slices explicitly provided to self.
            (i.e. don't get any special slices.)
        otherwise, create and include special slices in the result.
            see self.get_special_slices for details.
            if any special slices' names match slice names explicitly provided in self,
                use the explicitly-provided slice name instead.

        dims, ikeep: passed to self.get_special_slices, if creating any special slices.
            dims: {dims}
            ikeep: {ikeep}
        '''
        result = self.get_special_slices(dims=dims, ikeep=ikeep)
        result.update(self)
        return result

    @format_docstring(**_paramdocs_multi_slices, sub_ntab=1)
    def get_special_slices(self, ndim=UNSET, dims=UNSET, *, ikeep=UNSET):
        '''return dict of special slices to apply.

        ndim: UNSET or {ndim}
            UNSET --> use self.ndim
        dims: UNSET or {dims}
            UNSET --> use self.dims
        ikeep: UNSET or {ikeep}
            UNSET --> use self.ikeep

        Note that the naming convention for keys will be 'keep_'+'_'.join(unsliced_dims),
            e.g. 'keep_x', 'keep_x_y', or 'keep_x_y_z'.
            unsliced_dims will be in the same order as dims.

        returns dict of dicts, where each inner dict is instructions for applying a slice.
            result will be empty if ndim is None or if dims is None.
        '''
        if ndim is UNSET: ndim = self.ndim
        if dims is UNSET: dims = self.dims
        if ikeep is UNSET: ikeep = self.ikeep
        if ndim is None or dims is None:
            return dict()
        if ikeep is UNSET or ikeep is None:
            raise TypeError(f'ikeep should be an int or number between -1 and 1, but got ikeep={ikeep!r}')
        result = dict()
        for unsliced_dims in itertools.combinations(dims, ndim):
            dims_to_slice = tuple(dim for dim in dims if dim not in unsliced_dims)
            key = 'keep_'+'_'.join(unsliced_dims)
            result[key] = {dim: ikeep for dim in dims_to_slice}
        return result

    # # # DISPLAY # # #
    def _repr_contents(self):
        '''return list of strings to include in repr of self.'''
        contents = []
        if self.ndim is not None:
            contents.append(f'ndim={self.ndim}')
        if self.dims is not None:
            contents.append(f'dims={self.dims}')
        if self.ikeep != 0:
            contents.append(f'ikeep={self.ikeep}')
        return contents

    def __repr__(self):
        contents = self._repr_contents()
        contents.append(super().__repr__())
        contents_str = ', '.join(contents)
        return f'{type(self).__name__}({contents_str})'

    # # # EQUALITY # # #
    def __eq__(self, other):
        '''return whether self is equal to other.'''
        if not isinstance(other, type(self)):
            return False
        if self.ndim != other.ndim:
            return False
        if self.dims != other.dims:
            return False
        if self.ikeep != other.ikeep:
            return False
        return super().__eq__(other)

    def __ne__(self, other):
        '''return whether self is not equal to other.'''
        # (Defined in order to use __eq__ from this class instead of from super().)
        return not (self == other)
